import importlib
import os
import shutil
import tempfile
import inspect
import time
import threading
import queue

from jinja2 import Environment, FileSystemLoader

from naeural_core.business.base.web_app.base_web_app_plugin import BaseWebAppPlugin as BasePlugin
from naeural_core.utils.uvicorn_fast_api_ipc_manager import get_server_manager
from naeural_core.utils.fastapi_utils import PostponedRequest

#TODO: move __sign and __get_response from dauth_manager to base_web_app_plugin or fastapi_web_app or utils
#  all responses should contain the data from __get_response

__VER__ = '0.0.0'

DEFAULT_REQUEST_TIMEOUT = 2 * 60  # seconds
DEFAULT_MAX_INCOMING_REQUESTS = 1000
REQUEST_MONITOR_RESTART_MIN_BACKOFF_S = 0.5
REQUEST_MONITOR_RESTART_MAX_BACKOFF_S = 30.0
INCOMING_DROP_LOG_INTERVAL_S = 5.0
DEFAULT_MAX_REQUEST_MONITOR_RESTART_RETRIES = 10
DEFAULT_PROCESS_BUDGET_MS = 20
DEFAULT_MAX_INCOMING_PER_LOOP = 10
DEFAULT_MAX_POSTPONED_PER_LOOP = 10

_CONFIG = {
  **BasePlugin.CONFIG,
  'TUNNEL_ENGINE_ENABLED': True,

  'ENDPOINTS': [],
  'ASSETS': None,
  'JINJA_ARGS': {},
  'TEMPLATE': 'basic_server',

  'API_TITLE': None,  # default is plugin signature
  'API_SUMMARY': None,  # default is f"FastAPI created by {plugin signature}"
  'API_DESCRIPTION': None,  # default is plugin docstring

  'PAGES': [],
  'STATIC_DIRECTORY': 'assets',
  'DEFAULT_ROUTE': None,

  # In case of wrapped response, the response will be wrapped in a json with 2 keys:
  # 'result' and 'node_addr', where 'result' is the actual response and 'node_addr' is the node address
  # In case of raw response, the response will be the actual response provided by the endpoint method.
  # The default is 'WRAPPED'.
  'RESPONSE_FORMAT': 'WRAPPED',

  "LOG_REQUESTS": False,

  "REQUEST_TIMEOUT": DEFAULT_REQUEST_TIMEOUT,
  "MAX_INCOMING_REQUESTS": DEFAULT_MAX_INCOMING_REQUESTS,
  "MAX_REQUEST_MONITOR_RESTART_RETRIES": DEFAULT_MAX_REQUEST_MONITOR_RESTART_RETRIES,

  # Main-loop request budgeting. Set to 0 to disable.
  "PROCESS_BUDGET_MS": DEFAULT_PROCESS_BUDGET_MS,
  "MAX_INCOMING_PER_LOOP": DEFAULT_MAX_INCOMING_PER_LOOP,
  "MAX_POSTPONED_PER_LOOP": DEFAULT_MAX_POSTPONED_PER_LOOP,
  "FAIR_SCHEDULING": True,

  'PROCESS_DELAY': 0,

  # Profiling knobs (propagated into the generated uvicorn app).
  # `PROFILE_RATE` is the fraction of requests (0..1) that will be profiled.
  'PROFILE_RATE': 1.0,
  # Log aggregated timings every N seconds (0 disables periodic logs).
  'PROFILE_LOG_INTERVAL': 10 * 60,
  # If true, log each profiled request (high overhead).
  'PROFILE_LOG_PER_REQUEST': True,
  # If true, include detailed per-stage averages in periodic logs.
  'PROFILE_LOG_DETAILED': True,

  # No suppression of logs after interval
  "SUPRESS_LOGS_AFTER_INTERVAL": None,

  'VALIDATION_RULES': {
    **BasePlugin.CONFIG['VALIDATION_RULES']
  },
}


class FastApiWebAppPlugin(BasePlugin):
  """
  A plugin which exposes all of its methods marked with @endpoint through
  fastapi as http endpoints.

  The @endpoint methods can be triggered via http requests on the web server
  and will be processed as part of the business plugin loop.
  """

  CONFIG = _CONFIG
  
  @property
  def api_title(self):
    return repr(self.cfg_api_title or self.get_signature())
  
  @property
  def uvicorn_server_started(self):
    return self.__uvicorn_server_started

  def get_request_timeout(self):
    configured_request_timeout = self.cfg_request_timeout
    if not isinstance(configured_request_timeout, (int, float)):
      return None
    numeric_timeout = float(configured_request_timeout)
    if numeric_timeout <= 0:
      return None
    return numeric_timeout
  
  def get_max_incoming_requests(self):
    configured_value = self.cfg_max_incoming_requests
    if not isinstance(configured_value, int):
      return DEFAULT_MAX_INCOMING_REQUESTS
    if configured_value < 1:
      return 1
    return configured_value
  
  def get_max_request_monitor_restart_retries(self):
    configured_value = self.cfg_max_request_monitor_restart_retries
    if not isinstance(configured_value, int):
      return DEFAULT_MAX_REQUEST_MONITOR_RESTART_RETRIES
    if configured_value < 0:
      return 0
    return configured_value

  def get_process_budget_s(self):
    configured_value = self.cfg_process_budget_ms
    if not isinstance(configured_value, (int, float)):
      return None
    if configured_value <= 0:
      return None
    return float(configured_value) / 1000.0

  def get_max_incoming_per_loop(self):
    configured_value = self.cfg_max_incoming_per_loop
    if not isinstance(configured_value, int):
      return None
    if configured_value <= 0:
      return None
    return configured_value

  def get_max_postponed_per_loop(self):
    configured_value = self.cfg_max_postponed_per_loop
    if not isinstance(configured_value, int):
      return None
    if configured_value <= 0:
      return None
    return configured_value

  @staticmethod
  def cap_chunk_size(chunk_size):
    min_chunk_size = 128 * 1024
    default_chunk_size = 1024 * 1024
    max_chunk_size = 8 * 1024 * 1024
    if not isinstance(chunk_size, (int, float)):
      return default_chunk_size
    int_chunk_size = int(chunk_size)
    if int_chunk_size < min_chunk_size:
      return min_chunk_size
    if int_chunk_size > max_chunk_size:
      return max_chunk_size
    return int_chunk_size

  @staticmethod
  def endpoint(
      func=None, *,
      method="get",
      require_token=False,
      streaming_type=None,
      chunk_size=1024 * 1024
  ):
    """
    Decorator that marks a method as an HTTP endpoint with optional streaming support.

    Parameters
    ----------
    method : str
        HTTP method (e.g. "get", "post").
    require_token : bool
        Whether this endpoint should require a Bearer token.
    streaming_type : str or None
        Type of streaming: "upload", "download", or None for regular endpoints.
    chunk_size : int
        Size of chunks for streaming operations (default 1MB).
    """
    if func is None:
      def wrapper(func):
        return FastApiWebAppPlugin.endpoint(
          func, method=method, require_token=require_token,
          streaming_type=streaming_type, chunk_size=chunk_size
        )

      return wrapper

    func.__endpoint__ = True
    if isinstance(method, str):
      method = method.lower()
    func.__http_method__ = method
    func.__require_token__ = require_token
    func.__streaming_type__ = streaming_type
    func.__chunk_size__ = FastApiWebAppPlugin.cap_chunk_size(chunk_size)
    return func


  def get_web_server_path(self):
    return self.script_temp_dir

  def get_package_base_path(self, package_name):
    """
    Return the file path of an installed package parent directory.
    This method was copied from the _PluginsManagerMixin class from ratio1 SDK.

    Parameters
    ----------
    package_name : str
        The name of the installed package.

    Returns
    -------
    str
        The path to the package parent.
    """
    spec = importlib.util.find_spec(package_name)
    if spec is not None and spec.submodule_search_locations:
      return os.path.dirname(spec.submodule_search_locations[0])
    else:
      self.P("Package '{}' not found.".format(package_name), color='r')
    return None

  def validate_static_directory(self):
    """
    Validate the `STATIC_DIRECTORY` value.
    This will check if the value is trying to reference a restricted directory.
    """
    # Creating a temporary directory to check if the static directory is valid
    tmp_directory = tempfile.mkdtemp()
    static_directory = self.cfg_static_directory

    full_path = os.path.join(tmp_directory, static_directory)
    # Check if the full path is a subdirectory of the temporary directory
    if not full_path.startswith(tmp_directory):
      # If the full path is not a subdirectory of the temporary directory, it is invalid
      self.add_error(f"Invalid `STATIC_DIRECTORY`: {static_directory}")
    # endif path is restricted
    # Clean up the temporary directory
    os.rmdir(tmp_directory)
    return

  def validate_default_route(self):
    """
    Validate the `DEFAULT_ROUTE` value.
    It should be a valid URL path if specified.

    """
    default_route = self.cfg_default_route
    if default_route is None:
      # If not provided, return without validation
      return

    if not isinstance(default_route, str):
      self.add_error(f"Invalid `DEFAULT_ROUTE`: {default_route}. It should be a string")
      return

    if not default_route.startswith('/'):
      self.add_error(f"Invalid `DEFAULT_ROUTE`: {default_route}. It should start with '/'")
    # endif default route is invalid
    return

  def initialize_assets(self, src_dir, dst_dir, jinja_args):
    """
    Initialize and copy fastapi assets, expanding any jinja templates.
    All files from the source directory are copied to the
    destination directory with the following exceptions:
      - are symbolic links are ignored
      - files named ending with .jinja are expanded as jinja templates,
        .jinja is removed from the filename and the result copied to
        the destination folder.
    This maintains the directory structure of the source folder.
    In case src_dir is None, only the jinja templates are expanded.

    Parameters
    ----------
    src_dir: str or None, path to the source directory
    dst_dir: str, path to the destination directory
    jinja_args: dict, jinja keys to use while expanding the templates

    Returns
    -------
    None
    """
    self.prepared_env['PYTHONPATH'] = '.:' + os.getcwd() + ':' + self.prepared_env.get('PYTHONPATH', '')

    super(FastApiWebAppPlugin, self).initialize_assets(src_dir, dst_dir, jinja_args)

    package_base_path = self.get_package_base_path('naeural_core')
    if package_base_path is None:
      self.P("Skipping `main.py` rendering, package 'naeural_core' not found.", color='r')
      self.failed = True
      return
    # endif package base path not found
    static_directory = self.jinja_args.get('static_directory')

    if self.cfg_template is not None:
      env = Environment(loader=FileSystemLoader(package_base_path))

      # make sure static directory folder exists
      os.makedirs(self.os_path.join(dst_dir, static_directory), exist_ok=True)

      # Finally render main.py
      template_dir = self.os_path.join('naeural_core', 'business', 'base', 'uvicorn_templates')
      app_template = self.os_path.join(template_dir, f'{self.cfg_template}.j2')
      # env.get_template expects forward slashes, even on Windows.
      app_template = app_template.replace(os.sep, '/')
      app_template = env.get_template(app_template)
      rendered_content = app_template.render(jinja_args)

      with open(self.os_path.join(dst_dir, 'main.py'), 'w') as f:
        f.write(rendered_content)
    # endif render main.py

    # Here additional generic assets can be added if needed
    favicon_path = self.os_path.join(package_base_path, 'naeural_core', 'utils', 'web_app', 'favicon.ico')
    favicon_dst = self.os_path.join(dst_dir, static_directory, 'favicon.ico')
    if self.os_path.exists(favicon_path):
      self.P(f'Copying favicon from {favicon_path} to {favicon_dst}')
      os.makedirs(self.os_path.dirname(favicon_dst), exist_ok=True)
      shutil.copy2(favicon_path, favicon_dst)
    # endif favicon exists

    return

  def __register_custom_code_endpoint(self, endpoint_name, endpoint_method, endpoint_base64_code, endpoint_arguments):
    # First check that there is not any attribute with the same name
    import inspect
    existing_attribute_names = (name for name, _ in inspect.getmembers(self))
    if endpoint_name in existing_attribute_names:
      raise Exception("The endpoint name '{}' is already in use.".format(endpoint_name))

    custom_code_method, errors, warnings = self._get_method_from_custom_code(
      str_b64code=endpoint_base64_code,
      self_var='plugin',
      method_arguments=["plugin"] + endpoint_arguments
    )

    if errors is not None:
      raise Exception("The custom code failed with the following error: {}".format(errors))

    if len(warnings) > 0:
      self.P("The custom code generated the following warnings: {}".format("\n".join(warnings)))

    # Now register the custom code method as an endpoint
    import types
    setattr(
      self, endpoint_name, types.MethodType(
        FastApiWebAppPlugin.endpoint(
          custom_code_method, method=endpoint_method
        ), self
      )
    )
    return


  def _init_endpoints(self) -> None:
    """Enhanced to support streaming endpoints"""
    import inspect
    self._endpoints = {}
    jinja_args = []

    # Check for custom endpoints sent remotely.
    # These do not include any endpoints that serve static files or html pages.
    configured_endpoints = self.cfg_endpoints or []
    for dct_configured_endpoint in configured_endpoints:
      endpoint_name = dct_configured_endpoint.get('NAME', None)
      endpoint_method = dct_configured_endpoint.get('METHOD', "get")
      endpoint_base64_code = dct_configured_endpoint.get('CODE', None)
      endpoint_arguments = dct_configured_endpoint.get('ARGS', None)
      self.__register_custom_code_endpoint(
        endpoint_name=endpoint_name,
        endpoint_method=endpoint_method,
        endpoint_base64_code=endpoint_base64_code,
        endpoint_arguments=endpoint_arguments,
      )
    # endfor configured endpoints

    def _filter(obj):
      try:
        return inspect.ismethod(obj)
      except Exception as _:
        pass
      return False

    for name, method in inspect.getmembers(self, predicate=_filter):
      if not hasattr(method, '__endpoint__'):
        continue

      self._endpoints[name] = method
      http_method = method.__http_method__
      require_token = getattr(method, '__require_token__', False)
      streaming_type = getattr(method, '__streaming_type__', None)
      chunk_size = getattr(method, '__chunk_size__', 1024 * 1024)

      signature = inspect.signature(method)
      doc = method.__doc__ or ''
      has_kwargs = any([
        param.kind is inspect.Parameter.VAR_KEYWORD
        for param in signature.parameters.values()
      ])
      non_kw_params = [
        param for param in signature.parameters.values()
        if param.kind is not inspect.Parameter.VAR_KEYWORD
      ]
      all_params = [param.name for param in non_kw_params]
      all_args = [str(param) for param in non_kw_params]

      # Handle token parameter filtering as before
      if not require_token:
        args = all_args
        params = all_params
      else:
        if all_params[0] != 'token':
          raise ValueError(f"First parameter of method {name} must be 'token' if require_token is True.")
        params = all_params[1:]
        args = all_args[1:]

      jinja_args.append({
        'name': name,
        'method': http_method,
        'args': args,
        'params': params,
        'endpoint_doc': doc,
        'require_token': require_token,
        'has_kwargs': has_kwargs,
        'streaming_type': streaming_type,
        'chunk_size': chunk_size
      })

      streaming_info = f" (streaming: {streaming_type})" if streaming_type else ""
      str_function = f"{name}({', '.join(args)})"
      self.P(
        f"Registered endpoint {str_function} with method {http_method}{streaming_info}. Require token: {require_token}")

    self._node_comms_jinja_args = jinja_args
    return

  def on_init(self):
    self.__uvicorn_server_started = False
    # Register all endpoint methods.
    self._init_endpoints()
    
  

    # FIXME: move to setup_manager method
    self.manager_auth = b'abc'
    self._manager = get_server_manager(self.manager_auth)
    self._server_queue = self._manager.get_server_queue()
    self._client_queue = self._manager.get_client_queue()
    self.postponed_requests = self.deque()
    self._incoming_requests = self.deque()
    self._incoming_lock = threading.Lock()
    self._stop_request_monitor = threading.Event()
    self._request_monitor_thread = None
    self._request_monitor_last_restart_ts = 0.0
    self._request_monitor_backoff_s = 0.0
    self._request_monitor_restart_failures = 0
    self._last_incoming_drop_log_ts = 0.0

    self._profile_stats = {}
    self._profile_last_log_ts = self.time()

    self.P("manager address: {}".format(self._manager.address))
    _, self.manager_port = self._manager.address

    # Start the FastAPI app
    self.P('Starting FastAPI app...')
    super(FastApiWebAppPlugin, self).on_init()
    self._start_request_monitor_thread()
    return

  def get_deques_stats_str(self):
    with self._incoming_lock:
      incoming_len = len(self._incoming_requests)
    postponed_len = len(self.postponed_requests)
    server_queue_len = self._server_queue.qsize()
    return f"s:{server_queue_len} i:{incoming_len} p:{postponed_len}"

  def on_request(self, request):
    """
    Hook called when a new request arrives from the FastAPI side on the server_queue.

    Parameters
    ----------
    request : dict
      Object containing the following keys:
      - id: str - ID of the request
      - value: tuple - Method name and args for the endpoint handler
      - profile: dict or None - Profiling object
      - t_put_wall_ns: int or None - Wall-clock time when the request was enqueued
    """
    return

  def on_response(self, method: str, response: dict):
    """
    Hook method called before putting the response on the client_queue.

    Parameters
    ----------
    method : str
      Name of the HTTP method
    response : dict
      Object containing the following keys:
      - id: str - ID of the request
      - value: Any - Actual result of the endpoint handler
    """
    return

  def _start_request_monitor_thread(self):
    if self._request_monitor_thread is not None and self._request_monitor_thread.is_alive():
      return

    def _monitor():
      try:
        while not self._stop_request_monitor.is_set():
          try:
            request = self._server_queue.get(True, 0.25)
          except queue.Empty:
            continue
          except Exception as exc:
            self.P(f"Request monitor error: {exc}", color='r')
            continue
          try:
            value = request.get('value')
            is_profile_event = (
              isinstance(value, (list, tuple)) and len(value) > 0 and value[0] == "__profile__"
            )
            if not is_profile_event:
              t_start = time.perf_counter()
              self.on_request(request)
              timeout_s = self.get_request_timeout()
              if isinstance(timeout_s, (int, float)) and timeout_s > 0:
                elapsed = time.perf_counter() - t_start
                if elapsed > timeout_s:
                  self.P(
                    f"on_request exceeded timeout: {elapsed:.3f}s > {timeout_s:.3f}s",
                    color='y'
                  )
          except Exception:
            self.P(f"Exception in on_request:\n{self.get_exception()}", color='r')
          with self._incoming_lock:
            if len(self._incoming_requests) >= self.get_max_incoming_requests():
              now = time.time()
              if now - self._last_incoming_drop_log_ts >= INCOMING_DROP_LOG_INTERVAL_S:
                self._last_incoming_drop_log_ts = now
                self.P(
                  "Dropping incoming request: incoming queue is full.",
                  color='y'
                )
            else:
              self._incoming_requests.append(request)
      except Exception:
        self.P(f"Request monitor crashed:\n{self.get_exception()}", color='r')
        return

    self._request_monitor_thread = threading.Thread(
      target=_monitor,
      name="FastApiWebAppRequestMonitor",
      daemon=True
    )
    self._request_monitor_thread.start()
    return

  def create_postponed_request(self, solver_method, method_kwargs={}):
    """
    Create a postponed request to be processed by the plugin in the next loop.
    Parameters
    ----------
    solver_method : method
        The method that will solve the postponed request.
    method_kwargs : dict
        The keyword arguments to be passed to the solver_method.
    Returns
    -------
    res : PostponedRequest
        The postponed request object.
    """
    return PostponedRequest(
      solver_method=solver_method,
      method_kwargs=method_kwargs
    )

  def get_postponed_dict(self, request_id, request_value, endpoint_name):
    return {
      'id': request_id,
      'value': request_value,
      'endpoint_name': endpoint_name,
      'profile': None,
    }

  def parse_postponed_dict(self, request):
    return request['id'], request['value'], request['endpoint_name'], request.get('profile')

  def get_additional_fastapi_data(self):
    """
    Get additional data to be included in the FastAPI response.

    Returns
    -------
    dict - additional data to be included in the FastAPI response.
    """
    additional_env_vars = self.get_additional_env_vars()
    additional_env_vars = {
      (k.lower() if isinstance(k, str) else k): v for k, v in additional_env_vars.items()
    }
    return {
      # TO BE REMOVED maybe (they are also in the additional_data, but maybe the client expects these keys)
      'server_node_addr': self.e2_addr,
      'evm_network' : self.evm_network,
      **additional_env_vars
    }

  def __fastapi_process_response(self, response):
    if self.cfg_response_format == 'RAW':
      return response
    fastapi_additional = self.get_additional_fastapi_data()
    envelope = {
      'result': response,
      **fastapi_additional
    }
    if isinstance(response, dict):
      status_code = response.get('status_code')
      if status_code is not None:
        envelope['status_code'] = status_code
    return envelope

  def __fastapi_handle_response(self, id, value, profile, method):
    # TODO: add here message signing
    if self.cfg_log_requests:
      self.P(f"Request {id} for {method} processed.[{self.get_deques_stats_str()}]")
    response = {
      'id': id,
      'value': self.__fastapi_process_response(value)
    }
    if isinstance(profile, dict):
      response['profile'] = profile
    self.on_response(method, response)
    self._client_queue.put(response)
    return

  def _process_postponed_request(self, request):
    id, value, endpoint_name, profile = self.parse_postponed_dict(request)

    method = value.get_solver_method()
    kwargs = value.get_method_kwargs()

    try:
      if isinstance(profile, dict):
        self._profile_slice_start(profile, endpoint_name, t_put_wall_ns=None)
      value = method(**kwargs)
      if isinstance(profile, dict):
        self._profile_slice_end(profile)
    except Exception as exc:
      self.P(
        f'Exception occurred while processing postponed request for {endpoint_name} with method {method.__name__} '
        f'and args:\n{kwargs}\nException:\n{self.get_exception()}',
        color='r'
      )
      value = {
        'error': str(exc),
        'status_code': 500,
        'logged': True,
      }
      if isinstance(profile, dict):
        self._profile_slice_end(profile)

    if isinstance(value, PostponedRequest):
      postponed = self.get_postponed_dict(
        request_id=id,
        request_value=value,
        endpoint_name=endpoint_name
      )
      postponed['profile'] = profile
      return postponed

    self.__fastapi_handle_response(
      id=id,
      value=value,
      profile=profile,
      method=method,
    )
    return None

  def _process_incoming_request(self, request):
    id = request.get('id')
    value = request.get('value')
    profile = request.get('profile')
    t_put_wall_ns = request.get('t_put_wall_ns')

    if isinstance(value, (list, tuple)) and len(value) > 0 and value[0] == "__profile__":
      try:
        self._handle_profile_event(value[1])
      except Exception:
        self.P(f"Failed handling profile event:\n{self.get_exception()}", color='r')
      return None

    method = value[0]
    args = value[1:]

    try:
      if self.cfg_log_requests:
        self.P(f"Received request {id} for {method}.[{self.get_deques_stats_str()}]")
      endpoint = self._endpoints.get(method)
      if endpoint is None:
        raise ValueError(f"Endpoint '{method}' not found in registered endpoints.")
      has_kwargs = any(
        param.kind is inspect.Parameter.VAR_KEYWORD
        for param in inspect.signature(endpoint).parameters.values()
      )
      if has_kwargs and args and isinstance(args[-1], dict):
        kwargs = args[-1]
        args = args[:-1]
      else:
        kwargs = {}
      if isinstance(profile, dict):
        self._profile_slice_start(profile, method, t_put_wall_ns=t_put_wall_ns)
      value = endpoint(*args, **kwargs)
      if isinstance(profile, dict):
        self._profile_slice_end(profile)
    except Exception as exc:
      self.P("Exception occured while processing\n"
             "Request: {}\nArgs: {}\nException:\n{}".format(
                 method, args, self.get_exception()), color='r')
      status_code = 500
      if isinstance(exc, ValueError) and "not found" in str(exc):
        status_code = 404
      value = {
        'error': str(exc),
        'status_code': status_code,
        'logged': True,
      }
      if isinstance(profile, dict):
        self._profile_slice_end(profile)

    if isinstance(value, PostponedRequest):
      self.P(f"Postponing request {id} for {method}.")
      postponed = self.get_postponed_dict(
        request_id=id,
        request_value=value,
        endpoint_name=method
      )
      postponed['profile'] = profile
      return postponed

    self.__fastapi_handle_response(
      id=id,
      value=value,
      profile=profile,
      method=method
    )
    return None

  def _process(self):
    if not self._stop_request_monitor.is_set():
      if self._request_monitor_thread is None or not self._request_monitor_thread.is_alive():
        now = self.time()
        if self._request_monitor_backoff_s <= 0:
          self._request_monitor_backoff_s = REQUEST_MONITOR_RESTART_MIN_BACKOFF_S
        next_allowed = self._request_monitor_last_restart_ts + self._request_monitor_backoff_s
        if now >= next_allowed:
          max_retries = self.get_max_request_monitor_restart_retries()
          if self._request_monitor_restart_failures >= max_retries:
            msg = (
              "Request monitor thread failed to start after "
              f"{self._request_monitor_restart_failures} retries."
            )
            self.P(msg, color='r')
            raise RuntimeError(msg)
          self.P("Request monitor thread not running; restarting.", color='y')
          self._start_request_monitor_thread()
          self._request_monitor_last_restart_ts = now
          self._request_monitor_backoff_s = min(
            REQUEST_MONITOR_RESTART_MAX_BACKOFF_S,
            self._request_monitor_backoff_s * 2
          )
          self._request_monitor_restart_failures += 1
      else:
        self._request_monitor_backoff_s = 0.0
        self._request_monitor_restart_failures = 0
      # endif request monitor needs (re)starting
    # endif request monitor is not stopped

    new_postponed_requests = []
    process_budget_s = self.get_process_budget_s()
    deadline = None
    if isinstance(process_budget_s, (int, float)):
      deadline = time.perf_counter() + process_budget_s

    max_incoming = self.get_max_incoming_per_loop()
    max_postponed = self.get_max_postponed_per_loop()
    remaining_incoming = max_incoming
    remaining_postponed = max_postponed

    def _budget_exceeded():
      if deadline is None:
        return False
      return time.perf_counter() >= deadline

    def _can_take(remaining):
      return remaining is None or remaining > 0

    def _mark_incoming():
      nonlocal remaining_incoming
      if remaining_incoming is not None:
        remaining_incoming -= 1

    def _mark_postponed():
      nonlocal remaining_postponed
      if remaining_postponed is not None:
        remaining_postponed -= 1

    def _popleft_incoming():
      with self._incoming_lock:
        if not self._incoming_requests:
          return None
        return self._incoming_requests.popleft()

    def _popleft_postponed():
      if not self.postponed_requests:
        return None
      return self.postponed_requests.popleft()

    def _try_postponed():
      nonlocal remaining_postponed
      if not _can_take(remaining_postponed):
        return False
      request = _popleft_postponed()
      if request is None:
        return False
      postponed = self._process_postponed_request(request)
      if postponed is not None:
        new_postponed_requests.append(postponed)
      _mark_postponed()
      return True

    def _try_incoming():
      nonlocal remaining_incoming
      if not _can_take(remaining_incoming):
        return False
      request = _popleft_incoming()
      if request is None:
        return False
      postponed = self._process_incoming_request(request)
      if postponed is not None:
        self.postponed_requests.append(postponed)
      _mark_incoming()
      return True

    if self.cfg_fair_scheduling:
      next_is_postponed = True
      while True:
        if _budget_exceeded():
          break
        if (not _can_take(remaining_postponed)) and (
          not _can_take(remaining_incoming)
        ):
          break

        if next_is_postponed:
          processed = _try_postponed() or _try_incoming()
        else:
          processed = _try_incoming() or _try_postponed()

        if not processed:
          break
        next_is_postponed = not next_is_postponed
    else:
      while True:
        if _budget_exceeded() or (not _can_take(remaining_postponed)):
          break
        request = _popleft_postponed()
        if request is None:
          break
        postponed = self._process_postponed_request(request)
        if postponed is not None:
          new_postponed_requests.append(postponed)
        _mark_postponed()

      while True:
        if _budget_exceeded() or (not _can_take(remaining_incoming)):
          break
        request = _popleft_incoming()
        if request is None:
          break
        postponed = self._process_incoming_request(request)
        if postponed is not None:
          self.postponed_requests.append(postponed)
        _mark_incoming()
    # endif fair scheduling

    for request in new_postponed_requests:
      self.postponed_requests.append(request)

    self._maybe_log_profile_stats()
    super(FastApiWebAppPlugin, self)._process()
    return None

  def _profile_slice_start(self, profile, endpoint_name, t_put_wall_ns=None):
    profile.setdefault('endpoint', endpoint_name)
    if isinstance(t_put_wall_ns, int):
      profile.setdefault('t_put_wall_ns', t_put_wall_ns)
      profile['t_plugin_dequeue_wall_ns'] = time.time_ns()

    profile.setdefault('slice_count', 0)
    profile.setdefault('exec_total_ns', 0)
    t_slice_start = time.perf_counter_ns()
    profile.setdefault('t_endpoint_start_ns', t_slice_start)
    profile['t_slice_start_ns'] = t_slice_start
    return

  def _profile_slice_end(self, profile):
    slice_end = time.perf_counter_ns()
    profile['t_endpoint_end_ns'] = slice_end
    slice_start = profile.get('t_slice_start_ns')
    if isinstance(slice_start, int) and slice_end >= slice_start:
      profile['t_slice_end_ns'] = slice_end
      profile['slice_count'] = int(profile.get('slice_count', 0)) + 1
      profile['exec_total_ns'] = int(profile.get('exec_total_ns', 0)) + (slice_end - slice_start)
    return

  def _handle_profile_event(self, profile):
    if not isinstance(profile, dict):
      return
    if self.cfg_profile_rate <= 0:
      return

    endpoint = profile.get('endpoint') or 'unknown'
    def _delta_ms(a, b):
      if isinstance(a, int) and isinstance(b, int) and b >= a:
        return (b - a) / 1e6
      return None

    total_ms = _delta_ms(profile.get('t_http_start_ns'), profile.get('t_before_return_ns'))
    wait_ms = _delta_ms(profile.get('t_wait_start_ns'), profile.get('t_wait_end_ns'))
    queue_ms = _delta_ms(profile.get('t_put_wall_ns'), profile.get('t_plugin_dequeue_wall_ns'))
    exec_ms = None
    if isinstance(profile.get('exec_total_ns'), int):
      exec_ms = profile['exec_total_ns'] / 1e6

    def _from_start(target):
      start = profile.get('t_http_start_ns')
      if isinstance(start, int) and isinstance(target, int) and target >= start:
        return (target - start) / 1e6
      return None

    endpoint_start = profile.get('t_endpoint_start_ns')
    endpoint_end = profile.get('t_endpoint_end_ns') or profile.get('t_slice_end_ns')

    detailed = {
      'until_call_plugin_start': _from_start(profile.get('t_before_call_plugin_ns')),
      'until_put_start': _from_start(profile.get('t_put_start_ns')),
      'until_put_end': _from_start(profile.get('t_put_end_ns')),
      'until_wait_start': _from_start(profile.get('t_wait_start_ns')),
      'until_endpoint_start': _from_start(endpoint_start),
      # Here we use the slice end since the value should be from the last slice
      'until_endpoint_end': _from_start(endpoint_end),
      'until_wait_end': _from_start(profile.get('t_wait_end_ns')),
      'until_call_plugin_end': _from_start(profile.get('t_after_call_plugin_ns')),
      'until_return': _from_start(profile.get('t_before_return_ns')),
    }

    stats = self._profile_stats.get(endpoint)
    if stats is None:
      stats = {
        'n': 0,
        # short section
        'sum_total_ms': 0.0,
        'sum_wait_ms': 0.0,
        'sum_queue_ms': 0.0,
        'sum_exec_ms': 0.0,
        'max_total_ms': 0.0,
        # detailed section
        'sum_until_call_plugin_start': 0.0,
        'sum_until_put_start': 0.0,
        'sum_until_put_end': 0.0,
        'sum_until_wait_start': 0.0,
        'sum_until_endpoint_start': 0.0,
        'sum_until_endpoint_end': 0.0,
        'sum_until_wait_end': 0.0,
        'sum_until_call_plugin_end': 0.0,
        'sum_until_return': 0.0
      }
      self._profile_stats[endpoint] = stats

    stats['n'] += 1
    if total_ms is not None:
      stats['sum_total_ms'] += total_ms
      if total_ms > stats['max_total_ms']:
        stats['max_total_ms'] = total_ms
    if wait_ms is not None:
      stats['sum_wait_ms'] += wait_ms
    if queue_ms is not None:
      stats['sum_queue_ms'] += queue_ms
    if exec_ms is not None:
      stats['sum_exec_ms'] += exec_ms
    for key, val in detailed.items():
      sum_key = f"sum_{key}"
      if val is not None and sum_key in stats:
        stats[sum_key] += val

    if self.cfg_profile_log_per_request:
      total_ms = total_ms or 0.0
      wait_ms = wait_ms or 0.0
      queue_ms = queue_ms or 0.0
      exec_ms = exec_ms or 0.0
      base = f"<{endpoint}> timings(ms): t {total_ms:.2f} | w {wait_ms:.2f} | q {queue_ms:.2f} | e {exec_ms:.2f} | steps {profile.get('slice_count')}"
      if self.cfg_profile_log_detailed:
        detail = (
          f" | detail call_start={detailed.get('until_call_plugin_start') or 0:.2f} "
          f"put_s={detailed.get('until_put_start') or 0:.2f} put_e={detailed.get('until_put_end') or 0:.2f} "
          f"wait_s={detailed.get('until_wait_start') or 0:.2f} ep_s={detailed.get('until_endpoint_start') or 0:.2f} "
          f"ep_e={detailed.get('until_endpoint_end') or 0:.2f} wait_e={detailed.get('until_wait_end') or 0:.2f} "
          f"call_e={detailed.get('until_call_plugin_end') or 0:.2f} ret={detailed.get('until_return') or 0:.2f}"
        )
        self.P(base + detail)
      else:
        self.P(base)
    return

  def _maybe_log_profile_stats(self):
    if self.cfg_profile_rate <= 0:
      return
    interval = self.cfg_profile_log_interval
    if not isinstance(interval, (int, float)) or interval <= 0:
      return
    now = self.time()
    if now - self._profile_last_log_ts < interval:
      return
    self._profile_last_log_ts = now

    if not self._profile_stats:
      return

    parts = []
    for endpoint, stats in sorted(self._profile_stats.items()):
      n = max(int(stats.get('n', 0)), 1)
      avg_total = stats['sum_total_ms'] / n
      avg_wait = stats['sum_wait_ms'] / n
      avg_queue = stats['sum_queue_ms'] / n
      avg_exec = stats['sum_exec_ms'] / n
      max_total = stats.get('max_total_ms', 0.0)
      parts.append(f"{endpoint}: n {n} | t {avg_total:.2f} | w {avg_wait:.2f} | q {avg_queue:.2f} | e {avg_exec:.2f} | max_t {max_total:.2f}")
      if self.cfg_profile_log_detailed:
        def _avg(name):
          return stats.get(name, 0.0) / n
        parts.append(
          f"  detail: call_start={_avg('sum_until_call_plugin_start'):.2f} "
          f"put_s={_avg('sum_until_put_start'):.2f} put_e={_avg('sum_until_put_end'):.2f} "
          f"wait_s={_avg('sum_until_wait_start'):.2f} ep_s={_avg('sum_until_endpoint_start'):.2f} "
          f"ep_e={_avg('sum_until_endpoint_end'):.2f} wait_e={_avg('sum_until_wait_end'):.2f} "
          f"call_e={_avg('sum_until_call_plugin_end'):.2f} ret={_avg('sum_until_return'):.2f}"
        )
    self.P("FastAPI timings:\n" + "\n".join(parts))
    return

  def on_close(self):
    self._stop_request_monitor.set()
    if self._request_monitor_thread is not None:
      self._request_monitor_thread.join(timeout=1.0)
    self._manager.shutdown()
    super(FastApiWebAppPlugin, self).on_close()
    return

  def __get_uvicorn_process_args(self):
    return f"uvicorn --app-dir {self.script_temp_dir} main:app --host 0.0.0.0 --port {self.port}"

  def get_default_description(self):
    return self.__doc__
  
  
  def on_log_handler(self, text, key=None):
    super(FastApiWebAppPlugin, self).on_log_handler(text)
    STARTUP_MSG = "Uvicorn running on "
    if STARTUP_MSG in text:
      self.__uvicorn_server_started = True
      match = self.re.search(r"running on (.*?) \(Press", text)
      found = match.group(1) if match else ""
      self.P(f"Server {self.api_title} started on {found}", boxed=True)
    return

  def _setup_semaphore_env(self):
    """
    Expose the runtime FastAPI endpoint via semaphore for paired plugins.

    Uses the allocated runtime port (`self.port`), not the configured fallback
    (`self.cfg_port`), because web-app plugins may be rebound dynamically.
    """
    localhost_ip = None
    try:
      localhost_ip = self.log.get_localhost_ip()
    except Exception as ex:
      self.P("Failed to resolve localhost IP for API semaphore export: {}".format(ex), color='y')

    port = self.port
    if isinstance(localhost_ip, str) and localhost_ip.strip():
      self.semaphore_set_env('API_IP', localhost_ip)
    else:
      self.P("Skipping API_IP semaphore export: localhost IP unavailable", color='y')

    if port:
      self.semaphore_set_env('API_PORT', str(port))
      if isinstance(localhost_ip, str) and localhost_ip.strip():
        self.semaphore_set_env('API_URL', 'http://{}:{}'.format(localhost_ip, port))
      else:
        self.P("Skipping API_URL semaphore export: localhost IP unavailable", color='y')
    return

  @property
  def jinja_args(self):
    cfg_jinja_args = self.deepcopy(self.cfg_jinja_args)

    dct_pages = cfg_jinja_args.pop('html_files', self.cfg_pages)
    for page in dct_pages:
      page['method'] = 'get'

    static_directory = cfg_jinja_args.pop('static_directory', self.cfg_static_directory)
    default_route = self.cfg_default_route

    return {
      'static_directory': static_directory,
      'html_files': dct_pages,
      'manager_port': self.manager_port,
      'manager_auth': self.manager_auth,
      'api_title': self.api_title,
      'api_summary': repr(self.cfg_api_summary or f"Ratio1 WebApp created with {self.get_signature()} plugin"),
      'api_description': repr(self.cfg_api_description or self.get_default_description()),
      'api_version': repr(self.__version__),
      'node_comm_params': self._node_comms_jinja_args,
      'debug_web_app': self.cfg_debug_web_app,
      'default_route': default_route,
      'profile_rate': float(self.cfg_profile_rate or 0.0),
      'profile_log_per_request': bool(self.cfg_profile_log_per_request),
      'request_timeout': self.get_request_timeout(),
      'additional_fastapi_data': self.get_additional_fastapi_data(),
      **cfg_jinja_args,
    }

  def get_start_commands(self):
    super_start_commands = super(FastApiWebAppPlugin, self).get_start_commands()
    return [self.__get_uvicorn_process_args()] + super_start_commands



import os
import cv2
import numpy as np

from naeural_core import DecentrAIObject
from naeural_core import Logger

class Dataset(DecentrAIObject):
  def __init__(self, **kwargs):
    return

  def _is_image_path(self, path):
    return isinstance(path, str) and path.lower().endswith(('.jpg', '.jpeg', 'png', 'bmp'))

  def _load_image(self, path):
    if not isinstance(path, str):
      raise TypeError(f"Expected image path as str, got {type(path)}")
    img_bgr = cv2.imread(path)
    img_rgb = img_bgr[:, :, ::-1]
    img = np.ascontiguousarray(img_rgb)
    return img

  def load_images(self, paths):
    assert isinstance(paths, list)
    paths = [
      file for file in paths
      if self._is_image_path(file)
    ]
    check_files = [os.path.isfile(x) for x in paths]
    missing = [paths[i] for i, x in enumerate(check_files) if not x]
    if sum(check_files) != len(check_files):
      raise ValueError("Failed to find test images: {}".format(missing))

    imgs = []
    for path in paths:
      img = self._load_image(path)
      # name = os.path.basename(path)
      imgs.append(img)
    #endfor
    return imgs

  def load_images_from_folder(self, path):
    if not os.path.isdir(path):
      raise ValueError('Provided path is not a directory! Please check `{}`'.format(path))

    files = os.listdir(path)
    lst = self.load_images(files)
    return lst

  def load(self, items, default_type="IMG"):
    """
    Generic loader that preserves non-image items for structured tests.
    Returns the loaded items and an inferred input type (`IMG` or `STRUCT_DATA`).
    """
    if not isinstance(items, list):
      raise TypeError("Expected a list of items, got {}".format(type(items)))

    if len(items) == 0:
      return [], default_type

    if all(self._is_image_path(x) for x in items):
      return self.load_images(items), "IMG"

    return items, "STRUCT_DATA"


if __name__ == '__main__':
  log = Logger(
    lib_name='EE_TST',
    base_folder='.',
    app_folder='_local_cache',
    config_file='config_startup.txt',
    max_lines=1000,
    TF_KERAS=False
  )
  ds = Dataset()
  ds.test()



from collections import deque
from typing import Any, List

import numpy as np

from naeural_core import constants as ct
from naeural_core.data.base import DataCaptureThread
from naeural_core.data_structures import GeneralPayload

__VER__ = "0.1.0"

_CONFIG = {
  **DataCaptureThread.CONFIG,
  "LOOPBACK_QUEUE_SIZE": 32,
  "VALIDATION_RULES": {
    **DataCaptureThread.CONFIG["VALIDATION_RULES"],
    "LOOPBACK_QUEUE_SIZE": {
      "DESCRIPTION": "Maximum number of payloads cached in the loopback shared-memory queue.",
      "TYPE": "int",
      "MIN_VAL": 1,
      "MAX_VAL": 4096,
    },
  },
}


class LoopbackDataCapture(DataCaptureThread):
  CONFIG = _CONFIG


  def startup(self):
    super().startup()
    self._loopback_key = self._build_loopback_key(self.cfg_name)
    self._ensure_loopback_queue(self._loopback_key)
    self._metadata.update(pipeline_version=__VER__)
    return


  def connect(self):
    return True


  def data_step(self):
    queue = self._get_loopback_queue()
    if queue is None or len(queue) == 0:
      return

    pending_payloads = self._drain_queue(queue)
    if not pending_payloads:
      return

    self.P(f"Dequeued {len(pending_payloads)} payload(s) for stream '{self.cfg_name}'")
    inputs = self._build_inputs(pending_payloads)
    if inputs:
      self.P(f"Emitting {len(inputs)} input(s) downstream for stream '{self.cfg_name}'")
      self._add_inputs(inputs)
    else:
      self.P(f"No valid inputs built from payload batch on stream '{self.cfg_name}'", color='y')
    return


  def _release(self):
    return


  def _build_loopback_key(self, stream_name):
    return f"loopback_dct_{stream_name}"


  def _ensure_loopback_queue(self, key):
    existing = self.shmem.get(key)
    if isinstance(existing, deque):
      if existing.maxlen != self.cfg_loopback_queue_size:
        queue = deque(existing, maxlen=self.cfg_loopback_queue_size)
        self.shmem[key] = queue
      else:
        queue = existing
    else:
      queue = deque(maxlen=self.cfg_loopback_queue_size)
      self.shmem[key] = queue
    self._loopback_queue = queue
    return


  def _get_loopback_queue(self):
    queue = self.shmem.get(self._loopback_key)
    if not isinstance(queue, deque):
      self._ensure_loopback_queue(self._loopback_key)
      queue = self.shmem[self._loopback_key]
    return queue


  def _drain_queue(self, queue: deque):
    payloads: List[Any] = []
    while len(queue) > 0:
      payloads.append(queue.popleft())
    return payloads


  def _build_inputs(self, payloads: List[Any]):
    inputs = []
    for payload in payloads:
      if isinstance(payload, GeneralPayload):
        payload = payload.to_dict()

      if isinstance(payload, dict):
        img = payload.get("IMG")
        if isinstance(img, np.ndarray):
          inputs.append(
            self._new_input(img=img, struct_data=None, metadata=self._metadata.__dict__.copy())
          )
        elif img is not None:
          try:
            has_content = len(img) > 0  # type: ignore
          except TypeError:
            has_content = True
          if has_content:
            inputs.append(
              self._new_input(img=img, struct_data=None, metadata=self._metadata.__dict__.copy())
            )
          else:
            inputs.append(
              self._new_input(img=None, struct_data=payload, metadata=self._metadata.__dict__.copy())
            )
        else:
          inputs.append(
            self._new_input(img=None, struct_data=payload, metadata=self._metadata.__dict__.copy())
          )
      elif isinstance(payload, np.ndarray):
        inputs.append(
          self._new_input(img=payload, struct_data=None, metadata=self._metadata.__dict__.copy())
        )
      else:
        # fallback to structured data for unsupported payload types
        if payload is not None:
          inputs.append(
            self._new_input(img=None, struct_data=payload, metadata=self._metadata.__dict__.copy())
          )
    return inputs



# global dependencies
import gc
import random

import psutil
import torch as th
import pandas as pd
import numpy as np
import cv2
import os
import traceback

from time import sleep
from collections import OrderedDict

# local dependencies
from naeural_core import DecentrAIObject
from decentra_vision.draw_utils import DrawUtils
from naeural_core.serving.serving_manager import ServingManager
from naeural_core.serving.model_testing.utils import Dataset
from naeural_core.serving.model_testing.utils import ServingUtils


class Base(DecentrAIObject):
  def __init__(self,
               model_name,
               # Previously named test_files
               test_datasets,
               sleep_time=10,
               save_plots=True,
               show_plots=False,
               label_extension='json',
               nr_warmup=2,
               nr_predicts=20,
               inprocess=False,
               serving_manager=None,
               print_errors=True,
               subdir_keys=None,
               **kwargs):
    """

    :param model_name: tuple(CLASS, NAME)
      Passed to ServingManager.start_server server_name

    :param test_datasets: dict(dataset_name: [list of paths or list of raw inference items])
      Datasets to run inference on

      !! Previously named test_files - before non-image inference support was added.

    :param sleep_time: int
      Sleep time at startup

    :param save_plots: bool
      Whether to save ploted images to disk

    :param show_plots: bool
      Whether to show plotted images

    :param nr_warmup: int
      How many warmup steps to run

    :param nr_predicts: int
      How many predicts to run in order to get mean resource consumption

    :param inprocess: bool
      Whether to run inprocess serving processes

    :param serving_manager: ServingManager object

    :param print_errors: bool
      Whether to display inference errors

    :param subdir_keys: list
      keys used to better organize the saved results (mainly plots) in subdirectories

    :param kwargs:
    """
    super().__init__(**kwargs)
    self._model_name = model_name
    self._initial_test_datasets = test_datasets
    self._sleep_time = sleep_time
    self._save_plots = save_plots
    self._show_plots = show_plots
    # self._gpu_device = gpu_device
    self._serving_manager = serving_manager
    # if self._gpu_device is None:
    #   self._gpu_device = 1 if th.cuda.device_count() > 1 else 0
    self._nr_predicts = nr_predicts
    self._nr_warmup = nr_warmup
    self._inprocess = inprocess
    self._print_errors = print_errors
    self.subdir_keys = [] if subdir_keys is None else subdir_keys
    self.current_dct_params = {}

    if self._inprocess is True:
      self.P("WARNING !!! 'inprocess' is set to True which may yield invalid values because of the inability to do a complete cleanup", color='r')
    #
    # self._lst_imgs = None
    # self._lst_img_names = None

    self._dct_res = {
      'MODEL': [],
      'TIME': [],
      'TOTAL_MEM': [],
      'INITIALIZATION_MEM': [],
      'TIME_PER_INPUT': [],
      'NUMBER_OF_INPUTS': [],
      'SUCCESS': []
    }
    self._test_datasets = {}
    self._test_datasets_paths = {}
    self._dataset_types = {}
    self.label_extension = label_extension
    self._preds = OrderedDict()
    self._init()
    return

  def set_loop_stage(*args, **kwargs):
    return

  def _init(self):
    """Starting serving manager, starting utils. Loading data."""

    if self._serving_manager is None:
      self._serving_manager = ServingManager(
        shmem={},
        log=self.log,
        prefix_log='[TFMGR]',
        owner=self,
      )
    self._painter = DrawUtils(log=self.log)
    self._dataset = Dataset(log=self.log)
    self._serving_utils = ServingUtils(log=self.log)
    self._load_data()
    return

  def _load_data(self):
    """Loading testing dataset from indicated list of files or inference items"""

    self.log.P("Preparing data...", color='r')
    self.load_data()
    self.log.P("Done preparing data.")
    return

  def load_data(self):
    for dataset_name, dataset_paths in self._initial_test_datasets.items():
      if not isinstance(dataset_paths, list):
        dataset_paths = [dataset_paths]
      loaded_data, data_type = self._dataset.load(dataset_paths)
      self._test_datasets[dataset_name] = loaded_data
      self._dataset_types[dataset_name] = data_type
      if all(isinstance(dataset_path, str) for dataset_path in dataset_paths):
        self._test_datasets_paths[dataset_name] = [
          f'{os.path.splitext(dataset_path)[0]}.{self.label_extension}'
          for dataset_path in dataset_paths
        ]
      else:
        self._test_datasets_paths[dataset_name] = []
    # self._lst_imgs = list(dct.values())
    # self._lst_img_names = list(dct.keys())
    return

  def get_device_free_memory(
      self, device, device_id: int = None,
      show_logs: bool = True
  ):
    if device == 'cpu':
      res = psutil.virtual_memory().available
    else:
      gpus_info = self.log.gpu_info(show=show_logs, mb=True)
      if device_id is None:
        device_id = [
          i for i, gpu_info in enumerate(gpus_info)
          if gpus_info['NAME'] == th.cuda.get_device_name(device)
        ][0]
      # endif device_id not provided
      res = gpus_info[device_id]['FREE_MEM']
    return res

  def run(self, dct_test={}, dct_params={}, _lst_images=None, kill_serving_manager=True):
    """
    Execute inference testing
    Video file testing params:
      _lst_images = force the use of the given _lst_images
      kill_serving_manager = set to false in video file testing in order to not create a new manager for every movie batch

    Parameters
    ----------
    dct_test : dict
      Test specific parameters
    dct_params : dict
      Common parameters for all tests
    _lst_images : list
      List of images to use for testing
    kill_serving_manager : bool
      Whether to kill the serving manager after test

    Returns
    -------
    preds : list
      List of predictions
    """
    dct_aggregated = {
      **dct_params,
      **dct_test,
    }
    device_id = None
    preds = None
    default_device = dct_aggregated.get('DEFAULT_DEVICE', None)
    dataset_name = dct_aggregated.get('dataset_name')
    dataset_input_type = self.get_dataset_input_type(dataset_name) if dataset_name is not None else None
    input_type = dct_aggregated.get("INPUT_TYPE") or dataset_input_type or "IMG"

    if "MAX_BATCH_FIRST_STAGE" not in dct_aggregated:
      self.P("WARNING! Parameter 'MAX_BATCH_FIRST_STAGE' not set. Using default value 5", color='r')

    batch_first_stage = dct_aggregated.get("MAX_BATCH_FIRST_STAGE", 5)
    batch_second_stage = dct_aggregated.get("MAX_BATCH_SECOND_STAGE", 1)

    batch_size = max(batch_first_stage, batch_second_stage)

    run_dct = {
      'DEFAULT_DEVICE': None,
      'MODEL': None,
      'TIME': None,
      'TOTAL_MEM': None,
      'INITIALIZATION_MEM': None,
      'TIME_PER_INPUT': None,
      'NUMBER_OF_INPUTS': None,
      'SUCCESS': None,
      'PREDICTS': self._nr_predicts,
      'WARMUPS': self._nr_warmup,
      'INPUT_TYPE': input_type,
      **dct_aggregated
    }
    try:
      gpus_info = self.log.gpu_info(show=True, mb=True)
      if len(gpus_info) == 0:
        self.P(f"WARNING! No GPUs found. Running on CPU only.", color='r')
        default_device = "cpu"
      else:
        if default_device is not None:
          if default_device in ['gpu', 'cuda']:
            default_device = 'cuda:0'
          device_id = [
            i for i, gpu_info in enumerate(gpus_info)
            if gpus_info['NAME'] == th.cuda.get_device_name(default_device)
          ][0]
        else:
          default_device = "cuda:0"
          device_id = 0
        # endif default_device
      # endif no gpus
      run_dct['DEFAULT_DEVICE'] = default_device

      self.log.P("Running test for {}".format(dct_test), color='g')
      mem1 = self.get_device_free_memory(
        device=default_device,
        device_id=device_id
      )
      self.log.P(f"Free memory before start: {mem1}")

      str_model_name = self._serving_manager.start_server(
        server_name=self._model_name,
        inprocess=self._inprocess,
        upstream_config=dct_aggregated,
      )
      run_dct['MODEL'] = str_model_name

      mem15 = self.get_device_free_memory(
        device=default_device,
        device_id=device_id
      )
      initialization_mem = mem1 - mem15
      run_dct['INITIALIZATION_MEM'] = initialization_mem

      self.log.P("Model is ready. Check nvidia-smi in next {} seconds...".format(self._sleep_time))
      sleep(self._sleep_time)

      self.log.P("Running warm-up predict...")
      if _lst_images is None:
        _lst_images = self.get_inputs(dataset_name)

      no_imgs = len(_lst_images)
      _lst_images = (_lst_images * np.ceil(batch_size / no_imgs).astype(int))[:max(batch_size, no_imgs)]
      run_dct['NUMBER_OF_INPUTS'] = len(_lst_images)

      yolo_model_inputs = self._serving_utils.get_model_input(
        _lst_images,
        input_type=input_type
      )
      for _ in range(self._nr_warmup):
        preds = self._serving_manager.predict(self._model_name, yolo_model_inputs)
      # endfor

      if preds is None and self._nr_warmup > 0:
        raise Exception("Error raised in predict")

      self.log.reset_timers()

      self.log.P("Running batch predict...")
      yolo_model_inputs = self._serving_utils.get_model_input(
        _lst_images,
        input_type=input_type
      )
      preds = self._serving_manager.predict(self._model_name, yolo_model_inputs)

      if preds is None:
        raise Exception("Error raised in predict")

      for _ in range(self._nr_predicts - 1):
        random.shuffle(_lst_images)
        yolo_model_inputs = self._serving_utils.get_model_input(
          _lst_images,
          input_type=input_type
        )
        self._serving_manager.predict(self._model_name, yolo_model_inputs)
      # endfor

      self._preds[len(self._preds)] = preds

      mem2 = self.get_device_free_memory(
        device=default_device,
        device_id=device_id
      )
      self.log.P(f"Free memory after run: {mem2}")
      total_model_mem = mem1 - mem2
      run_dct['TOTAL_MEM'] = total_model_mem
      self.log.P("Model memory {:.0f} MB".format(total_model_mem), color='m')

      self.log.P("Showing results & timers...", color='g')
      self.log.show_timers(title="on {} for '{}' with {}".format(
          self.log.get_machine_name(),
          self._model_name, dct_test
        ),
        div=max(len(_lst_images), 1)
      )

      if dataset_name is not None:
        plot_kwargs = {**dct_test, **dct_params, "INPUT_TYPE": input_type}
        if input_type == "IMG":
          self.plot(**plot_kwargs)
        elif self._show_plots or self._save_plots:
          self.log.P(f"Skipping plot for dataset '{dataset_name}' with non-image input type '{input_type}'")

        res = self.score(**plot_kwargs)
        if res is not None:
          run_dct['SCORE'] = res
        # endif res None
      # endif dataset available

      if not self._show_plots and not self._save_plots:
        self.log.P("Skipped showing images & inferences.")
      # endif

      mem3 = self.get_device_free_memory(
        device=default_device,
        device_id=device_id
      )
      self.log.P(f"Free memory after shutdown: {mem3}")

      if kill_serving_manager:
        self._serving_manager.stop_server(server_name=self._model_name)

      timer_prefix = 'local_pred_' if self._inprocess else 'remote_pred_'
      run_dct['TIME'] = self.log.get_timer(timer_prefix + str_model_name.upper())['MEAN']
      run_dct['TIME_PER_INPUT'] = self.log.get_timer(timer_prefix + str_model_name.upper())['MEAN'] / len(_lst_images)
      run_dct['SUCCESS'] = True
    except:
      run_dct['SUCCESS'] = False
      if self._print_errors:

        info = traceback.format_exc()
        self.log.P("Inference error: {}".format(info), color='r')
    # endtry

    for k, v in run_dct.items():
      if k not in self._dct_res:
        self._dct_res[k] = []
      self._dct_res[k].append(v)

    gc.collect()
    th.cuda.empty_cache()
    return preds

  def get_results(self):
    df = pd.DataFrame(self._dct_res)
    return df

  def show_results(self):
    df = self.get_results()
    self.log.set_nice_prints()
    self.log.P("Results on {}:\n{}".format(self.log.get_machine_name(), df))
    return

  @property
  def _testing_subfolder_path(self):
    return self._testing_subfolder_path_helper()

  def _testing_subfolder_path_helper(self, skip_subdirs=False):
    res = os.path.join('testing', self.log.file_prefix, self.__name__, self._model_name)
    if not skip_subdirs and len(self.subdir_keys) > 0:
      values = [self.current_dct_params.get(key) for key in self.subdir_keys if key in self.current_dct_params]
      str_values = [str(value) for value in values]
      res = os.path.join(res, *str_values)
    # endif subdir_keys
    return res

  def save_results(self):
    df = self.get_results()
    df_name = 'results.csv'
    self.log.save_dataframe(df, df_name, folder='output',
                            subfolder_path=self._testing_subfolder_path_helper(skip_subdirs=True))

  def run_tests(self, lst_tests, dct_params, save_results=True):
    """
    Run a series of tests over all datasets
    Parameters
    ----------
    lst_tests : list of dicts
      Each dict contains a test configuration
    dct_params : dict
      Common parameters for all tests
    save_results : bool
      Whether to save results to disk

    Returns
    -------
    pd.DataFrame
      Dataframe with results
    """
    # TODO: big ambiguity between lists of test dicts and dct_params
    for dataset_name in self.get_dataset_names():
      for dct_test in lst_tests:
        test_params = {'dataset_name': dataset_name, **dct_test}
        self.current_dct_params = test_params
        self.run(test_params, dct_params)
    # endfor
    self.show_results()
    if save_results:
      self.save_results()
    return self.get_results()

  def plot(self, dataset_name, **kwargs):
    """Draw plots specific to each kind of model"""

    raise NotImplementedError()
    return

  def score(self, dataset_name, **kwargs):
    """Score computing for specific model"""

    return None

  def plot_video(self, cv2_video_writer, lst_imgs, preds, **kwargs):
    """Based on a lst of imgs and preds save a video"""
    raise NotImplementedError()
    return

  def get_inputs(self, dataset_name):
    return self._test_datasets[dataset_name]

  def get_images(self, dataset_name):
    return self._test_datasets[dataset_name]

  def get_dataset_input_type(self, dataset_name):
    return self._dataset_types.get(dataset_name, "IMG")

  def get_dataset_names(self):
    return list(self._test_datasets.keys())

  def get_dataset_output_folder(self, dataset_name):
    return os.path.join(self._testing_subfolder_path, dataset_name)

  def get_dataset_labels_path(self, dataset_name):
    return self._test_datasets_paths[dataset_name]

  def get_preds(self, idx=None):
    preds = self._preds
    if idx is not None:
      preds = self._preds.get(idx, None)
    return preds

  def get_last_preds(self):
    dct_preds = self.get_preds()
    keys = list(dct_preds.keys())
    preds = dct_preds[keys[-1]]
    return preds

  def get_model_name(self):
    return self._model_name

  def run_video_test(self, video_fn, test_params={}, flip_video=False, movie_batch_size=50, plot_video_kwargs={}):
    """

    :param video_fn:
    :param flip_video:
      False - nothing happens
      Rotate code - passed to cv2.rotate
    :param movie_batch_size:
    :return:
    """
    cap = cv2.VideoCapture(video_fn)
    while not cap.isOpened():
      cap = cv2.VideoCapture(video_fn)

    frames = []
    fps = cap.get(cv2.CAP_PROP_FPS)
    pos_frame = cap.get(cv2.CAP_PROP_POS_FRAMES)
    while True:
      flag, frame = cap.read()
      if flip_video is not False:
        frame = cv2.rotate(frame, flip_video)
      frame_rgb = frame[:, :, ::-1]
      frame_rgb = np.ascontiguousarray(frame_rgb)
      frames.append(frame_rgb)

      if cap.get(cv2.CAP_PROP_POS_FRAMES) == cap.get(cv2.CAP_PROP_FRAME_COUNT):
          # If the number of captured frames is equal to the total number of frames,
          # we stop
        break

    output_video_file = '{}_{}'.format(self.log.file_prefix, os.path.split(video_fn)[-1])

    output_video_folder = os.path.join(
      self.log.get_output_folder(),
      'video_testing',
    )
    if not os.path.exists(output_video_folder):
      os.makedirs(output_video_folder)

    output_video_fn = os.path.join(output_video_folder, output_video_file)

    cv2_video_writer = cv2.VideoWriter(
      output_video_fn,
      cv2.VideoWriter_fourcc(*"XVID"),
      fps,
      (frames[0].shape[1], frames[0].shape[0])
    )

    for batch_id in range(len(frames) // movie_batch_size):
      self.P("Processing movie batch {}/{}".format(batch_id + 1, len(frames) // movie_batch_size))
      batch = frames[batch_id * movie_batch_size: (batch_id + 1) * movie_batch_size]
      batch_results = self.run(dct_test={**test_params, "MAX_BATCH_FIRST_STAGE": 5},
                               _lst_images=batch, kill_serving_manager=False)
      self.plot_video(
        cv2_video_writer=cv2_video_writer,
        lst_imgs=batch,
        preds=batch_results,
        **plot_video_kwargs
      )
    cv2_video_writer.release()

    a = 0


if __name__ == '__main__':
  from naeural_core import Logger
  from plugins.serving.testing.th_yolov5l6s.data import files as TEST_FILES

  log = Logger('MPTF', base_folder='.', app_folder='_solisbox', TF_KERAS=False)

  # defining script constants
  GPU_DEVICE = 1 if th.cuda.device_count() > 1 else 0
  TESTS = [
    # {
    #   'USE_AMP'         : False,
    #   'USE_FP16'        : False,
    #   'GPU_PREPROCESS'  : False,
    # },

    {
      'USE_AMP': True,
      'USE_FP16': False,
      'GPU_PREPROCESS': True,
      'DEFAULT_DEVICE': 'cuda:' + str(GPU_DEVICE)
    },

    # {
    #   'USE_AMP'         : True,
    #   'USE_FP16'        : False,
    #   'GPU_PREPROCESS'  : True,
    # },

    # {
    #   'USE_AMP'         : False,
    #   'USE_FP16'        : True,
    #   'GPU_PREPROCESS'  : True,
    # },
  ]
  COVERED = 'eff_det0'
  MODEL_NAME = ('th_y5l6s', 'default')

  test = Base(
    log=log,
    model_name=MODEL_NAME,
    test_datasets=TEST_FILES,
    gpu_device=GPU_DEVICE,
  )

  dct_params = {
    "NEW_PARAM": 0,
    "COVERED_SERVERS": ['th_y5l6', COVERED],
    "NMS_CONF_THR": 0.05,
    "WARMUP_BATCH": len(test.get_images())
  }
  test.run_tests(
    lst_tests=TESTS,
    dct_params=dct_params
  )



import gc
import traceback
import platform
import subprocess
import shutil

class _GPUMixin(object):
  """
  Mixin for GPU functionalities that are attached to `libraries.logger.Logger`.

  This mixin cannot be instantiated because it is built just to provide some additional
  functionalities for `libraries.logger.Logger`

  In this mixin we can use any attribute/method of the Logger.

  * Obs: This mixin uses also attributes/methods of `_MachineInfoMixin`:
    - self.get_machine_memory
    - self.get_avail_memory
  """

  def __init__(self):
    super(_GPUMixin, self).__init__()

    try:
      from ratio1.logging.logger_mixins.machine_mixin import _MachineMixin
    except ModuleNotFoundError:
      raise ModuleNotFoundError("Cannot use _GPUMixin without having _MachineMixin")

    self._done_first_smi_error = False
    self._nvml_initialized = False
    return

  @staticmethod
  def clear_gpu_memory():
    try:
      import tensorflow as tf
      import gc
      gc.collect()
      tf.keras.backend.clear_session()
    except:
      pass
    try:
      import torch as th
      import gc
      gc.collect()
      th.cuda.empty_cache()
      th.cuda.clear_memory_allocated()
    except:
      pass

  @staticmethod
  def get_gpu_memory_map():
    import subprocess
    """Get the current gpu usage.

    Returns
    -------
    usage: dict
        Keys are device ids as integers.
        Values are memory usage as integers in MB.
    """
    result = subprocess.check_output(
      [
        'nvidia-smi', '--query-gpu=memory.used',
        '--format=csv,nounits,noheader'
      ])
    result = result.decode('utf-8')
    # Convert lines into a dictionary
    gpu_memory = [int(x) for x in result.strip().split('\n')]
    gpu_memory_map = dict(zip(range(len(gpu_memory)), gpu_memory))
    return gpu_memory_map

  def skip_gpu_info_check(self):
    return vars(self).get('_GPUMixin__no_gpu_avail', False)

  def _get_processes_by_uuid_via_nvidia_smi(self, mb: bool):
    """
    Returns:
      processes_by_uuid: dict[str, list[dict]]   # uuid -> [{'PID':..., 'ALLOCATED_MEM':...}, ...]
    Notes:
      nvidia-smi used_gpu_memory is in MiB when nounits is used in CSV output (common behavior).
    """
    processes_by_uuid = {}

    smi = shutil.which("nvidia-smi")
    if not smi:
      return processes_by_uuid  # keep behavior: processes empty if unavailable

    try:
      cmd = [
        smi,
        "--query-compute-apps=gpu_uuid,pid,used_gpu_memory",
        "--format=csv,noheader,nounits",
      ]
      out = subprocess.run(cmd, capture_output=True, text=True, timeout=1.5)
      if out.returncode != 0:
        return processes_by_uuid

      for line in out.stdout.splitlines():
        line = line.strip()
        if not line:
          continue
        # Expect: "<uuid>, <pid>, <used_gpu_memory>"
        parts = [p.strip() for p in line.split(",")]
        if len(parts) < 3:
          continue

        uuid = parts[0]
        try:
          pid = int(parts[1])
        except Exception:
          continue

        used_mib = None
        try:
          # used_gpu_memory in MiB, nounits -> numeric
          used_mib = float(parts[2])
        except Exception:
          used_mib = None

        # Convert to bytes to reuse your existing MB/GB conversion logic
        used_bytes = None if used_mib is None else int(used_mib * 1024 * 1024)

        dct_proc_info = {"PID": pid}
        dct_proc_info["ALLOCATED_MEM"] = round(
          (used_bytes / 1024 ** (2 if mb else 3)) if used_bytes is not None else 0.0,
          2
        )

        processes_by_uuid.setdefault(uuid, []).append(dct_proc_info)
      # endfor lines

    except Exception:
      # Any failure here should degrade gracefully to empty processes
      return {}

    return processes_by_uuid

  def _get_uuid_by_index(self, timeout=1.5):
    import subprocess, shutil
    smi = shutil.which("nvidia-smi")
    if not smi:
      return {}

    out = subprocess.run(
      [smi, "--query-gpu=index,uuid", "--format=csv,noheader,nounits"],
      capture_output=True, text=True, timeout=timeout
    )
    if out.returncode != 0:
      return {}

    d = {}
    for line in out.stdout.splitlines():
      line = line.strip()
      if not line:
        continue
      idx_s, uuid = [p.strip() for p in line.split(",", 1)]
      d[int(idx_s)] = uuid
    return d

  def gpu_info(self, show=False, mb=False, current_pid=False):
    """
    Collects GPU info. Must have torch installed & non-mandatory nvidia-smi

    Parameters
    ----------
    show : bool, optional
      show data as gathered. The default is False.
    mb : bool, optional
      collect memory in MB otherwise in GB. The default is False.
    current_pid: bool, optional
      return data only for GPUs used by current process or all if current process does
    not use GPU


    Returns
    -------
    lst_inf : list of dicts
      all GPUs info from CUDA:0 to CUDA:n.

    """

    def _main_func():
      if self.skip_gpu_info_check():
        return []

      try:
        # first get name
        import torch as th
        import os
        processes_by_uuid = self._get_processes_by_uuid_via_nvidia_smi(mb=mb)
      except:
        self.P("ERROR: `gpu_info` call failed - PyTorch probably is not installed:\n{}".format(
          traceback.format_exc())
        )
        self.__no_gpu_avail = True
        return None

      nvsmires, device_props, dct_proc_info = None, None, None
      pynvml = None

      try:
        import pynvml  # provided by nvidia-ml-py
        if not self._nvml_initialized:
          pynvml.nvmlInit()
          self._nvml_initialized = True
        pynvml_avail = True
      except Exception:
        pynvml_avail = False

      if not pynvml_avail:
        self.__no_gpu_avail = True

      lst_inf = []
      # now we iterate all devices
      n_gpus = th.cuda.device_count()
      if n_gpus > 0:
        gc.collect()
        th.cuda.empty_cache()
      current_pid_has_usage = False
      current_pid_gpus = []

      try:
        for device_id in range(n_gpus):
          dct_device = {}
          device_props = th.cuda.get_device_properties(device_id)
          dct_device['NAME'] = device_props.name
          dct_device['TOTAL_MEM'] = round(
            device_props.total_memory / 1024 ** (2 if mb else 3),
            2
          )
          mem_total = None
          mem_allocated = None
          gpu_used = None
          gpu_temp = None
          gpu_temp_max = None
          fan_speed, fan_speed_unit = -1, "N/A"
          if pynvml_avail:
            # --- get an NVML handle that matches torch's CUDA device ordering when possible ---
            uuid_by_index = self._get_uuid_by_index()
            # inside your for device_id in range(n_gpus):
            uuid = uuid_by_index.get(device_id)
            if uuid:
              handle = pynvml.nvmlDeviceGetHandleByUUID(uuid)
            else:
              handle = pynvml.nvmlDeviceGetHandleByIndex(device_id)

            # --- memory (NVML returns bytes) ---
            if True:
              mem = pynvml.nvmlDeviceGetMemoryInfo(handle)
              mem_total = round(mem.total / 1024 ** (2 if mb else 3), 2)
              mem_allocated = round(mem.used / 1024 ** (2 if mb else 3), 2)
            # endif mem

            # --- utilization ---
            if True:
              util = pynvml.nvmlDeviceGetUtilizationRates(handle)
              gpu_used = util.gpu  # int percent
            # endif gpu_used

            # --- temperature ---
            if True:
              gpu_temp = pynvml.nvmlDeviceGetTemperature(handle, pynvml.NVML_TEMPERATURE_GPU)
            # endif gpu_temp

            # mimic DeviceQuery's gpu_temp_max_threshold (commonly shutdown threshold)
            if True:
              gpu_temp_max = None
              try:
                if hasattr(pynvml, "NVML_TEMPERATURE_THRESHOLD_SHUTDOWN"):
                  gpu_temp_max = pynvml.nvmlDeviceGetTemperatureThreshold(
                    handle, pynvml.NVML_TEMPERATURE_THRESHOLD_SHUTDOWN
                  )
                elif hasattr(pynvml, "NVML_TEMPERATURE_THRESHOLD_SLOWDOWN"):
                  # fallback if shutdown isn't exposed in older wrappers
                  gpu_temp_max = pynvml.nvmlDeviceGetTemperatureThreshold(
                    handle, pynvml.NVML_TEMPERATURE_THRESHOLD_SLOWDOWN
                  )
              except Exception:
                gpu_temp_max = None
            # endif gpu_temp_max

            # --- fan ---
            if True:
              fan_speed, fan_speed_unit = -1, "N/A"
              try:
                fan_speed = pynvml.nvmlDeviceGetFanSpeed(handle)  # percent
                fan_speed_unit = "%"
              except Exception:
                fan_speed, fan_speed_unit = -1, "N/A"
            # endif fan

            # --- processes (keep your existing shape) ---
            if True:
              processes = []
              try:
                # Use NVML only to get the UUID for this device, then map from nvidia-smi results.
                uuid = None
                try:
                  uuid = pynvml.nvmlDeviceGetUUID(handle)
                except Exception:
                  uuid = None

                if uuid and uuid in processes_by_uuid:
                  processes = processes_by_uuid[uuid]

                  # Preserve your existing "current pid GPU usage" behavior
                  for p in processes:
                    if p.get("PID") == os.getpid():
                      current_pid_has_usage = True
                      current_pid_gpus.append(device_id)
              except Exception:
                processes = []
              # endtry processes
            # endif processes

            dct_device["PROCESSES"] = processes
            dct_device["USED_BY_PROCESS"] = device_id in current_pid_gpus
          else:
            str_os = platform.platform()
            ## check if platform is Tegra and record
            if 'tegra' in str_os.lower():
              # we just record the overall fre memory
              mem_total = self.get_machine_memory(gb=not mb)
              mem_allocated = mem_total  - self.get_avail_memory(gb=not mb)
              gpu_used = 1
              gpu_temp = 1
              gpu_temp_max = 100
              if not self._done_first_smi_error and nvsmires is not None:
                self.P("Running `gpu_info` on Tegra platform: {}".format(nvsmires), color='r')
                self._done_first_smi_error = True
            elif not self._done_first_smi_error:
              str_log = "ERROR: Please make sure you have both pytorch and pynvml in order to monitor the GPU"
              str_log += "\nError info: pynvml_avail={}, nvsmires={}".format(pynvml_avail, nvsmires)
              self.P(str_log)
              self._done_first_smi_error = True
          #endif
          dct_device['ALLOCATED_MEM'] = mem_allocated
          dct_device['FREE_MEM'] = -1
          if all(x is not None for x in [mem_total, mem_allocated]):
            dct_device['FREE_MEM'] = round(mem_total - mem_allocated, 2)
          dct_device['MEM_UNIT'] = 'MB' if mb else 'GB'
          dct_device['GPU_USED'] = gpu_used
          dct_device['GPU_TEMP'] = gpu_temp
          dct_device['GPU_TEMP_MAX'] = gpu_temp_max
          dct_device['GPU_FAN_SPEED'] = fan_speed
          dct_device['GPU_FAN_SPEED_UNIT'] = fan_speed_unit

          lst_inf.append(dct_device)
        #end for all devices
      except Exception as e:
        self.P("gpu_info exception for device_id {}: {}\n devicepros: {}\n nvsmires: {}\n dct_proc_info: {}".format(
          device_id, e,
          device_props, nvsmires, dct_proc_info), color='r'
        )

      if show:
        self.P("GPU information for {} device(s):".format(len(lst_inf)), color='y')
        for dct_gpu in lst_inf:
          for k, v in dct_gpu.items():
            self.P("  {:<14} {}".format(k + ':', v), color='y')

      if current_pid and current_pid_has_usage:
        return [lst_inf[x] for x in current_pid_gpus]
      else:
        return lst_inf
    #enddef

    # if multiple threads call at the same time the method `log.gpu_info`, then `pynvml_avail` will be False
    #   (most probably due to the queries that are performed using nvidia_smi)
    with self.managed_lock_resource('gpu_info'):
      res = _main_func()
    # endwith lock
    return res

  
  def get_gpu_info(self, device_id=0, mb=False):
    """
    Returns GPU info for a specific device
    """
    res = {}
    if not self.skip_gpu_info_check():
      gpu_info = self.gpu_info()
      if gpu_info is not None and len(gpu_info) > 0 and device_id < len(gpu_info):
        res = gpu_info[device_id]
    return res
  
  
  def get_gpu_total_mem(self, device_id=0, mb=False):
    """
    Returns total memory of a specific GPU device
    """
    res = 0
    if not self.skip_gpu_info_check():
      device_info = self.get_gpu_info(device_id=device_id, mb=mb)
      if device_info is not None and len(device_info) > 0:
        res = device_info['TOTAL_MEM']
    return res
  
  
  def get_gpu_free_mem(self, device_id=0, mb=False):
    """
    Returns free memory of a specific GPU device
    """
    res = 0
    if not self.skip_gpu_info_check():
      device_info = self.get_gpu_info(device_id=device_id, mb=mb)
      if device_info is not None and len(device_info) > 0:
        res = device_info['FREE_MEM']
    return res
  
  
  def get_gpu_name(self, device_id=0):
    """
    Returns name of a specific GPU device
    """
    res = 'N/A'
    if not self.skip_gpu_info_check():
      device_info = self.get_gpu_info(device_id=device_id)
      if device_info is not None and len(device_info) > 0:
        res = device_info['NAME']
    return res
  
  
  def get_gpu_fan_speed(self, device_id=0):
    """
    Returns fan speed of a specific GPU device
    """
    res = 'N/A'
    if not self.skip_gpu_info_check():
      device_info = self.get_gpu_info(device_id=device_id)
      if device_info is not None and len(device_info) > 0:
        res = device_info.get('GPU_FAN_SPEED')
    return res
    


if __name__ == '__main__':
  print(_GPUMixin.get_gpu_memory_map())



import os
import sys
import threading
import time
from collections import deque
from datetime import datetime as dt
from copy import deepcopy

REPO_ROOT = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
if REPO_ROOT not in sys.path:
  sys.path.insert(0, REPO_ROOT)

from naeural_core import Logger
from naeural_core.constants import HB, PAYLOAD_DATA
from naeural_core.main.net_mon import NetworkMonitor, NETMON_MUTEX, UNUSEFULL_HB_KEYS


class DummyEpochManager:
  def register_data(self, addr, data):
    return


class UnsafeNetworkMonitor(NetworkMonitor):
  """
  Reintroduce the old in-place mutation behavior to reproduce the race.
  """
  def unsafe_register_heartbeat(self, addr, data, sleep_between=0.0005):
    # mimic NetworkMonitor.register_heartbeat + old __register_heartbeat body
    data[HB.RECEIVED_TIME] = dt.now().strftime(HB.TIMESTAMP_FORMAT)

    if HB.ENCODED_DATA in data:
      str_data = data.pop(HB.ENCODED_DATA)
      dct_hb = self.log.decompress_text(str_data)
      data = {
        **data,
        **dct_hb,
      }

    __addr_no_prefix = self._NetworkMonitor__remove_address_prefix(addr)

    for key_to_delete in UNUSEFULL_HB_KEYS:
      data.pop(key_to_delete, None)

    with self.log.managed_lock_resource(NETMON_MUTEX):
      if __addr_no_prefix not in self._NetworkMonitor__network_heartbeats:
        self._NetworkMonitor__network_heartbeats[__addr_no_prefix] = deque(maxlen=self.HB_HISTORY)
      # append shared dict
      self._NetworkMonitor__network_heartbeats[__addr_no_prefix].append(data)
      # widen race window
      time.sleep(sleep_between)
      # mutate stored dict in place (old behavior)
      self._NetworkMonitor__maybe_register_hb_pipelines(addr, data)
      # mutate previous hb in place (old behavior)
      if len(self._NetworkMonitor__network_heartbeats[__addr_no_prefix]) >= 2:
        self._NetworkMonitor__pop_repeating_info_from_heartbeat(
          self._NetworkMonitor__network_heartbeats[__addr_no_prefix][-2]
        )


def _make_logger():
  return Logger(
    lib_name="NMON_TEST",
    base_folder=".",
    app_folder="_local_cache",
    no_folders_no_save=True,
    max_lines=50,
    DEBUG=False,
  )


def _make_hb(addr, extra_keys=500):
  hb = {
    HB.EE_ADDR: addr,
    HB.CURRENT_TIME: dt.now().strftime(HB.TIMESTAMP_FORMAT),
    PAYLOAD_DATA.EE_TIMESTAMP: dt.utcnow().strftime(HB.TIMESTAMP_FORMAT),
    PAYLOAD_DATA.EE_TIMEZONE: "UTC",
    f"{HB.PREFIX_EE_NODETAG}DC": "TEST_DC",
    f"{HB.PREFIX_EE_NODETAG}REG": "EU",
  }
  # inflate dict to slow deepcopy and increase race probability
  for i in range(extra_keys):
    hb[f"K{i}"] = i
  return hb


def _run_race(netmon, addr, writer_fn, duration_sec=2.0):
  stop = threading.Event()
  error = {"exc": None}

  def reader():
    while not stop.is_set():
      try:
        netmon.get_network_node_tags(addr)
      except Exception as exc:  # capture RuntimeError
        error["exc"] = exc
        stop.set()
        return

  def writer():
    while not stop.is_set():
      hb = _make_hb(addr)
      writer_fn(addr, hb)

  t_r = threading.Thread(target=reader, daemon=True)
  t_w = threading.Thread(target=writer, daemon=True)
  t_r.start()
  t_w.start()

  stop.wait(duration_sec)
  stop.set()
  t_r.join(timeout=1.0)
  t_w.join(timeout=1.0)

  return error["exc"]


def test_race_repro_old_behavior():
  log = _make_logger()
  netmon = UnsafeNetworkMonitor(
    log=log,
    node_name="test_node",
    node_addr="aixp_test_node",
    epoch_manager=DummyEpochManager(),
  )
  addr = "aixp_test_node"

  # seed with one heartbeat
  netmon.unsafe_register_heartbeat(addr, _make_hb(addr))

  exc = _run_race(
    netmon=netmon,
    addr=addr,
    writer_fn=netmon.unsafe_register_heartbeat,
    duration_sec=2.0,
  )

  if exc is None:
    raise AssertionError(
      "Expected RuntimeError from in-place mutation race, but none occurred."
    )
  if "dictionary changed size during iteration" not in str(exc):
    raise AssertionError(f"Unexpected exception: {exc}")


def test_race_fixed_behavior():
  log = _make_logger()
  netmon = NetworkMonitor(
    log=log,
    node_name="test_node",
    node_addr="aixp_test_node",
    epoch_manager=DummyEpochManager(),
  )
  addr = "aixp_test_node"

  # seed with one heartbeat
  netmon.register_heartbeat(addr, _make_hb(addr))

  exc = _run_race(
    netmon=netmon,
    addr=addr,
    writer_fn=netmon.register_heartbeat,
    duration_sec=2.0,
  )

  if exc is not None:
    raise AssertionError(f"Did not expect exception with fixed behavior: {exc}")


def main():
  # Ensure we are importing from the local workspace, not a site-packages install.
  import naeural_core as _nc
  if not os.path.abspath(_nc.__file__).startswith(REPO_ROOT):
    raise RuntimeError(f"naeural_core import is not local: {_nc.__file__}")
  tests = [
    test_race_repro_old_behavior,
    test_race_fixed_behavior,
  ]
  failures = 0
  for test in tests:
    name = test.__name__
    try:
      test()
      print(f"PASS: {name}")
    except Exception as exc:
      failures += 1
      print(f"FAIL: {name}: {exc}")
  if failures:
    raise SystemExit(1)


if __name__ == "__main__":
  main()



#!/usr/bin/env python3
"""
Passive NET_MON_01 compression probe using the same zlib+base64 codec as
heartbeat ENCODED_DATA.

This captures real mainnet NET_MON_01 payloads, stores full raw samples for
evidence, and estimates the size reduction for:
1. CURRENT_NETWORK-only compression
2. Heartbeat-style body compression of all non-EE_* payload fields
"""

from __future__ import annotations

import argparse
import base64
import glob
import hashlib
import json
import zlib
from dataclasses import dataclass
from datetime import datetime, timezone
from pathlib import Path
from threading import Lock
from time import sleep, time
from typing import Any

try:
  from xperimental.payloads_tests.sdk_bandwidth_capture import (
    PassiveBandwidthMqttSession,
    compact_json_text,
    human_bytes,
    message_class,
    pct,
    percentile,
    render_sender_label,
    utc_now_iso,
  )
except ModuleNotFoundError:
  from sdk_bandwidth_capture import (  # type: ignore
    PassiveBandwidthMqttSession,
    compact_json_text,
    human_bytes,
    message_class,
    pct,
    percentile,
    render_sender_label,
    utc_now_iso,
  )


DEFAULT_OUTPUT_DIR = "xperimental/payloads_tests/evidence/netmon_compression"
CURRENT_NETWORK_VERSION = "zlib_b64_v1"
NETMON_BODY_VERSION = "v2"
TARGET_MESSAGE_CLASS = "payload:NET_MON_01"


def hb_like_compress_text(text: str) -> str:
  return base64.b64encode(zlib.compress(text.encode("utf-8"), level=9)).decode("utf-8")


def compact_json_bytes(value: Any) -> int:
  return len(compact_json_text(value).encode("utf-8"))


def latest_bandwidth_summary(default_glob: str) -> Path | None:
  matches = sorted(glob.glob(default_glob))
  if not matches:
    return None
  return Path(matches[-1])


@dataclass
class ProbeArtifacts:
  samples_jsonl: Path
  summary_json: Path
  results_md: Path
  metadata_json: Path


class JsonlWriter:
  def __init__(self, path: Path):
    self.path = path
    self.path.parent.mkdir(parents=True, exist_ok=True)
    self._handle = self.path.open("w", encoding="utf-8")
    self._lock = Lock()

  def write(self, row: dict[str, Any]) -> None:
    with self._lock:
      self._handle.write(json.dumps(row, ensure_ascii=True) + "\n")
      self._handle.flush()

  def close(self) -> None:
    with self._lock:
      self._handle.close()


class NetmonCompressionRecorder:
  def __init__(self, samples_path: Path):
    self.writer = JsonlWriter(samples_path)
    self.lock = Lock()
    self.count = 0
    self.first_recv_ts = None
    self.last_recv_ts = None

  def record(self, *, comm_type: str, topic: str, payload_bytes: bytes, log) -> None:
    recv_ts = time()
    raw_wire_bytes = len(payload_bytes)
    try:
      raw_text = payload_bytes.decode("utf-8")
      raw_obj = json.loads(raw_text)
    except Exception:
      return

    path = raw_obj.get("EE_PAYLOAD_PATH") or [None, None, None, None]
    signature = path[2] if isinstance(path, list) and len(path) > 2 else None
    event_type = raw_obj.get("EE_EVENT_TYPE")
    if message_class(event_type, signature, raw_obj) != TARGET_MESSAGE_CLASS:
      return

    sender_addr = raw_obj.get("EE_SENDER")
    sender_id = raw_obj.get("EE_ID")
    current_network = raw_obj.get("CURRENT_NETWORK") or {}
    current_network_text = compact_json_text(current_network)
    current_network_raw_bytes = len(current_network_text.encode("utf-8"))
    current_network_zlib_bytes = len(zlib.compress(current_network_text.encode("utf-8"), level=9))
    current_network_encoded = hb_like_compress_text(current_network_text)
    current_network_encoded_bytes = len(current_network_encoded.encode("utf-8"))

    current_network_only_payload = dict(raw_obj)
    current_network_only_payload["CURRENT_NETWORK"] = current_network_encoded
    current_network_only_payload["CURRENT_NETWORK_VERSION"] = CURRENT_NETWORK_VERSION
    current_network_only_payload_bytes = compact_json_bytes(current_network_only_payload)

    body = {
      key: value for key, value in raw_obj.items()
      if not str(key).startswith("EE_")
    }
    body_text = compact_json_text(body)
    body_raw_bytes = len(body_text.encode("utf-8"))
    body_encoded = hb_like_compress_text(body_text)
    body_encoded_bytes = len(body_encoded.encode("utf-8"))
    hb_like_payload = {
      key: value for key, value in raw_obj.items()
      if str(key).startswith("EE_")
    }
    hb_like_payload["NETMON_VERSION"] = NETMON_BODY_VERSION
    hb_like_payload["ENCODED_DATA"] = body_encoded
    hb_like_payload_bytes = compact_json_bytes(hb_like_payload)

    row = {
      "sample_index": None,
      "recv_ts": recv_ts,
      "recv_iso": datetime.fromtimestamp(recv_ts, tz=timezone.utc).isoformat(timespec="seconds"),
      "comm_type": comm_type,
      "topic": topic,
      "sender": render_sender_label(sender_id, sender_addr),
      "sender_id": sender_id,
      "sender_addr": sender_addr,
      "raw_wire_bytes": raw_wire_bytes,
      "raw_compact_bytes": compact_json_bytes(raw_obj),
      "payload_sha1": hashlib.sha1(payload_bytes).hexdigest()[:12],
      "current_network_raw_bytes": current_network_raw_bytes,
      "current_network_zlib_bytes": current_network_zlib_bytes,
      "current_network_encoded_bytes": current_network_encoded_bytes,
      "current_network_only_payload_bytes": current_network_only_payload_bytes,
      "current_network_only_reduction_bytes": raw_wire_bytes - current_network_only_payload_bytes,
      "current_network_only_reduction_pct": pct(raw_wire_bytes - current_network_only_payload_bytes, raw_wire_bytes),
      "hb_like_body_raw_bytes": body_raw_bytes,
      "hb_like_body_encoded_bytes": body_encoded_bytes,
      "hb_like_payload_bytes": hb_like_payload_bytes,
      "hb_like_reduction_bytes": raw_wire_bytes - hb_like_payload_bytes,
      "hb_like_reduction_pct": pct(raw_wire_bytes - hb_like_payload_bytes, raw_wire_bytes),
      "body_keys": sorted(body.keys()),
      "top_level_keys": sorted(raw_obj.keys()),
      "raw_payload": raw_obj,
    }

    with self.lock:
      self.count += 1
      row["sample_index"] = self.count
      if self.first_recv_ts is None:
        self.first_recv_ts = recv_ts
      self.last_recv_ts = recv_ts
    self.writer.write(row)

  def close(self) -> None:
    self.writer.close()


def build_artifacts(output_dir: Path) -> ProbeArtifacts:
  ts = datetime.now(timezone.utc).strftime("%Y%m%dT%H%M%S%z")
  output_dir.mkdir(parents=True, exist_ok=True)
  base = output_dir / f"{ts}_netmon_compression"
  return ProbeArtifacts(
    samples_jsonl=base.with_suffix(".jsonl"),
    summary_json=Path(str(base) + "_summary.json"),
    results_md=Path(str(base) + "_results.md"),
    metadata_json=Path(str(base) + "_metadata.json"),
  )


def render_table(headers: list[str], rows: list[list[Any]]) -> str:
  lines = [
    "| " + " | ".join(headers) + " |",
    "| " + " | ".join(["---"] * len(headers)) + " |",
  ]
  for row in rows:
    lines.append("| " + " | ".join(str(cell) for cell in row) + " |")
  return "\n".join(lines)


def analyze_samples(
  samples_jsonl: Path,
  metadata_path: Path,
  results_md: Path,
  summary_json: Path,
  bandwidth_summary_path: Path | None,
  args: argparse.Namespace,
) -> dict[str, Any]:
  rows = []
  with samples_jsonl.open("r", encoding="utf-8") as handle:
    for line in handle:
      rows.append(json.loads(line))

  metadata = json.loads(metadata_path.read_text(encoding="utf-8"))
  raw_sizes = [row["raw_wire_bytes"] for row in rows]
  current_network_only_sizes = [row["current_network_only_payload_bytes"] for row in rows]
  hb_like_sizes = [row["hb_like_payload_bytes"] for row in rows]
  current_network_reduction_pcts = [row["current_network_only_reduction_pct"] for row in rows]
  hb_like_reduction_pcts = [row["hb_like_reduction_pct"] for row in rows]
  current_network_field_reduction_pcts = [
    pct(row["current_network_raw_bytes"] - row["current_network_encoded_bytes"], row["current_network_raw_bytes"])
    for row in rows
  ]

  totals = {
    "sample_count": len(rows),
    "raw_wire_bytes": sum(raw_sizes),
    "current_network_only_payload_bytes": sum(current_network_only_sizes),
    "hb_like_payload_bytes": sum(hb_like_sizes),
    "current_network_only_reduction_bytes": sum(row["current_network_only_reduction_bytes"] for row in rows),
    "hb_like_reduction_bytes": sum(row["hb_like_reduction_bytes"] for row in rows),
    "current_network_raw_bytes": sum(row["current_network_raw_bytes"] for row in rows),
    "current_network_encoded_bytes": sum(row["current_network_encoded_bytes"] for row in rows),
    "hb_like_body_raw_bytes": sum(row["hb_like_body_raw_bytes"] for row in rows),
    "hb_like_body_encoded_bytes": sum(row["hb_like_body_encoded_bytes"] for row in rows),
  }

  bandwidth_projection = None
  if bandwidth_summary_path is not None and bandwidth_summary_path.exists():
    bandwidth_summary = json.loads(bandwidth_summary_path.read_text(encoding="utf-8"))
    total_raw_bytes = bandwidth_summary["totals"]["raw_bytes"]
    netmon_row = next(
      (row for row in bandwidth_summary["by_class"] if row["message_class"] == TARGET_MESSAGE_CLASS),
      None,
    )
    if netmon_row is not None:
      netmon_raw_bytes = netmon_row["total_bytes"]
      avg_hb_like_reduction_pct = (
        totals["hb_like_reduction_bytes"] / totals["raw_wire_bytes"] * 100.0
        if totals["raw_wire_bytes"] > 0 else 0.0
      )
      avg_current_network_only_reduction_pct = (
        totals["current_network_only_reduction_bytes"] / totals["raw_wire_bytes"] * 100.0
        if totals["raw_wire_bytes"] > 0 else 0.0
      )
      bandwidth_projection = {
        "bandwidth_summary_path": str(bandwidth_summary_path),
        "ten_min_total_raw_bytes": total_raw_bytes,
        "ten_min_netmon_raw_bytes": netmon_raw_bytes,
        "hb_like_saved_bytes": int(round(netmon_raw_bytes * avg_hb_like_reduction_pct / 100.0)),
        "current_network_only_saved_bytes": int(round(netmon_raw_bytes * avg_current_network_only_reduction_pct / 100.0)),
      }
      bandwidth_projection["hb_like_saved_share_of_total"] = pct(
        bandwidth_projection["hb_like_saved_bytes"],
        total_raw_bytes,
      )
      bandwidth_projection["current_network_only_saved_share_of_total"] = pct(
        bandwidth_projection["current_network_only_saved_bytes"],
        total_raw_bytes,
      )

  results = {
    "scope": {
      "samples_jsonl": str(samples_jsonl),
      "summary_json": str(summary_json),
      "results_md": str(results_md),
      "sample_count": len(rows),
      "capture_seconds": metadata["duration_seconds"],
      "stop_reason": metadata["stop_reason"],
    },
    "totals": totals,
    "percentiles": {
      "raw_wire_bytes_p50": percentile(raw_sizes, 50),
      "raw_wire_bytes_p95": percentile(raw_sizes, 95),
      "current_network_only_reduction_pct_p50": percentile([int(round(x * 10)) for x in current_network_reduction_pcts], 50) / 10.0,
      "hb_like_reduction_pct_p50": percentile([int(round(x * 10)) for x in hb_like_reduction_pcts], 50) / 10.0,
    },
    "bandwidth_projection": bandwidth_projection,
    "largest_samples": sorted(rows, key=lambda row: row["raw_wire_bytes"], reverse=True)[:10],
  }
  summary_json.write_text(json.dumps(results, indent=2, ensure_ascii=True), encoding="utf-8")

  avg_current_network_only_reduction_pct = (
    totals["current_network_only_reduction_bytes"] / totals["raw_wire_bytes"] * 100.0
    if totals["raw_wire_bytes"] > 0 else 0.0
  )
  avg_hb_like_reduction_pct = (
    totals["hb_like_reduction_bytes"] / totals["raw_wire_bytes"] * 100.0
    if totals["raw_wire_bytes"] > 0 else 0.0
  )
  current_network_field_reduction_pct = (
    (totals["current_network_raw_bytes"] - totals["current_network_encoded_bytes"]) / totals["current_network_raw_bytes"] * 100.0
    if totals["current_network_raw_bytes"] > 0 else 0.0
  )

  scenario_table = render_table(
    ["scenario", "total bytes", "avg bytes/sample", "reduction vs raw"],
    [
      [
        "raw NET_MON_01",
        human_bytes(totals["raw_wire_bytes"]),
        int(totals["raw_wire_bytes"] / len(rows)) if rows else 0,
        "0.0%",
      ],
      [
        "CURRENT_NETWORK compressed",
        human_bytes(totals["current_network_only_payload_bytes"]),
        int(totals["current_network_only_payload_bytes"] / len(rows)) if rows else 0,
        f"{avg_current_network_only_reduction_pct:.1f}%",
      ],
      [
        "hb-style NET_MON body",
        human_bytes(totals["hb_like_payload_bytes"]),
        int(totals["hb_like_payload_bytes"] / len(rows)) if rows else 0,
        f"{avg_hb_like_reduction_pct:.1f}%",
      ],
    ],
  )

  field_table = render_table(
    ["field/body", "raw bytes", "hb-compatible encoded bytes", "reduction"],
    [
      [
        "CURRENT_NETWORK only",
        human_bytes(totals["current_network_raw_bytes"]),
        human_bytes(totals["current_network_encoded_bytes"]),
        f"{current_network_field_reduction_pct:.1f}%",
      ],
      [
        "all non-EE_* NET_MON fields",
        human_bytes(totals["hb_like_body_raw_bytes"]),
        human_bytes(totals["hb_like_body_encoded_bytes"]),
        f"{pct(totals['hb_like_body_raw_bytes'] - totals['hb_like_body_encoded_bytes'], totals['hb_like_body_raw_bytes'])}%",
      ],
    ],
  )

  sample_table = render_table(
    ["sample", "sender", "raw bytes", "CURRENT_NETWORK reduction", "hb-like reduction"],
    [
      [
        row["sample_index"],
        row["sender"],
        human_bytes(row["raw_wire_bytes"]),
        f"{row['current_network_only_reduction_pct']}%",
        f"{row['hb_like_reduction_pct']}%",
      ]
      for row in results["largest_samples"][:10]
    ],
  )

  projection_lines = ["- no 10-minute raw-bandwidth summary was available for extrapolation"]
  if bandwidth_projection is not None:
    projection_lines = [
      f"- source 10-minute summary: `{bandwidth_projection['bandwidth_summary_path']}`",
      f"- `NET_MON_01` raw bytes in that 10-minute run: `{human_bytes(bandwidth_projection['ten_min_netmon_raw_bytes'])}`",
      f"- projected save with CURRENT_NETWORK-only compression: `{human_bytes(bandwidth_projection['current_network_only_saved_bytes'])}` "
      f"({bandwidth_projection['current_network_only_saved_share_of_total']}% of the full 10-minute raw sample)",
      f"- projected save with hb-style NET_MON body compression: `{human_bytes(bandwidth_projection['hb_like_saved_bytes'])}` "
      f"({bandwidth_projection['hb_like_saved_share_of_total']}% of the full 10-minute raw sample)",
    ]

  md = f"""# NET_MON Compression Probe Results

## Scope
- script path: `xperimental/payloads_tests/netmon_compression_probe.py`
- capture command: `python3 xperimental/payloads_tests/netmon_compression_probe.py --seconds {args.seconds} --target-count {args.target_count}`
- sample count: `{len(rows)}`
- capture window: `{metadata['start_iso']}` to `{metadata['end_iso']}` ({metadata['duration_seconds']:.1f}s)
- stop reason: `{metadata['stop_reason']}`

## Codec
- heartbeat-compatible codec: `zlib.compress(level=9)` + `base64.b64encode`
- reference implementation: `ratio1/logging/logger_mixins/general_serialization_mixin.py`
- note: payload-size estimates keep the outer JSON envelope and approximate post-change wire bytes by compact JSON serialization; `EE_SIGN` / `EE_HASH` values would change, but their lengths stay effectively constant

## Executive Summary
- Average raw NET_MON_01 size in this sample was `{human_bytes(totals['raw_wire_bytes'] / len(rows)) if rows else '0B'}`.
- Compressing `CURRENT_NETWORK` alone reduced payload bytes by `{avg_current_network_only_reduction_pct:.1f}%` across the sampled NET_MON payloads.
- Compressing the full non-`EE_*` NET_MON body in a heartbeat-style `ENCODED_DATA` envelope reduced payload bytes by `{avg_hb_like_reduction_pct:.1f}%`.
- The raw `CURRENT_NETWORK` field itself shrank by `{current_network_field_reduction_pct:.1f}%` when encoded with the heartbeat codec.

## Scenario Totals
{scenario_table}

## Field/Body Compression
{field_table}

## Largest Samples
{sample_table}

## Reduction Spread
- `CURRENT_NETWORK`-only reduction p50: `{percentile([int(round(x * 10)) for x in current_network_reduction_pcts], 50) / 10.0}%`
- `CURRENT_NETWORK`-only reduction p95: `{percentile([int(round(x * 10)) for x in current_network_reduction_pcts], 95) / 10.0}%`
- hb-style NET_MON-body reduction p50: `{percentile([int(round(x * 10)) for x in hb_like_reduction_pcts], 50) / 10.0}%`
- hb-style NETMON-body reduction p95: `{percentile([int(round(x * 10)) for x in hb_like_reduction_pcts], 95) / 10.0}%`
- raw `CURRENT_NETWORK` field reduction p50: `{percentile([int(round(x * 10)) for x in current_network_field_reduction_pcts], 50) / 10.0}%`

## Projection Onto The 10-Minute Raw Bandwidth Run
{chr(10).join(projection_lines)}

## Artifacts
- samples jsonl: `{samples_jsonl}`
- summary json: `{summary_json}`
- results md: `{results_md}`

## Verification
- command: `python3 xperimental/payloads_tests/netmon_compression_probe.py --seconds {args.seconds} --target-count {args.target_count}`
  result: `pass`
  evidence: `Captured {len(rows)} NET_MON_01 payloads and computed hb-style compression estimates`
"""
  results_md.write_text(md, encoding="utf-8")
  return results


def parse_args() -> argparse.Namespace:
  parser = argparse.ArgumentParser(description="Passive NET_MON_01 compression probe.")
  parser.add_argument("--seconds", type=int, default=180, help="Capture duration in seconds.")
  parser.add_argument("--target-count", type=int, default=30, help="Number of NET_MON_01 payloads to sample.")
  parser.add_argument("--network", default="mainnet", help="EVM network for the SDK session.")
  parser.add_argument("--output-dir", default=DEFAULT_OUTPUT_DIR, help="Directory for persistent evidence artifacts.")
  parser.add_argument("--session-name", default="netmon-compression-probe", help="SDK session alias.")
  parser.add_argument("--bandwidth-summary", help="Existing raw-bandwidth summary JSON for extrapolation.")
  parser.add_argument("--analyze-only", action="store_true", help="Skip live capture and analyze an existing samples JSONL.")
  parser.add_argument("--samples-jsonl", help="Existing NET_MON sample JSONL to analyze.")
  parser.add_argument("--metadata-file", help="Existing NET_MON sample metadata JSON to analyze.")
  parser.add_argument("--results-md", help="Explicit markdown output path.")
  parser.add_argument("--summary-json", help="Explicit summary json output path.")
  parser.add_argument("--silent", action="store_true", help="Reduce SDK logs during live capture.")
  return parser.parse_args()


def run_capture(args: argparse.Namespace) -> ProbeArtifacts:
  artifacts = build_artifacts(Path(args.output_dir))
  recorder = NetmonCompressionRecorder(artifacts.samples_jsonl)
  start_ts = time()
  metadata = {
    "start_ts": start_ts,
    "start_iso": utc_now_iso(),
    "network": args.network,
    "session_name": args.session_name,
    "stop_reason": "running",
  }
  artifacts.metadata_json.write_text(json.dumps(metadata, indent=2), encoding="utf-8")

  session = None
  try:
    session = PassiveBandwidthMqttSession(
      name=args.session_name,
      silent=args.silent,
      auto_configuration=True,
      run_dauth=True,
      use_home_folder=False,
      local_cache_base_folder=".",
      evm_network=args.network,
      raw_record_callback=recorder.record,
    )
    deadline = start_ts + args.seconds
    while time() < deadline and recorder.count < args.target_count:
      sleep(0.25)
    metadata["stop_reason"] = "target-count" if recorder.count >= args.target_count else "time-window"
  finally:
    metadata["end_ts"] = time()
    metadata["end_iso"] = utc_now_iso()
    metadata["duration_seconds"] = round(metadata["end_ts"] - start_ts, 3)
    metadata["sample_count"] = recorder.count
    artifacts.metadata_json.write_text(json.dumps(metadata, indent=2), encoding="utf-8")
    recorder.close()
    if session is not None:
      session.close(wait_close=True)
  return artifacts


def main() -> int:
  args = parse_args()
  bandwidth_summary = Path(args.bandwidth_summary) if args.bandwidth_summary else latest_bandwidth_summary(
    "xperimental/payloads_tests/evidence/raw_bandwidth/*_mainnet_bandwidth_summary.json"
  )

  if args.analyze_only:
    if not args.samples_jsonl:
      raise SystemExit("--samples-jsonl is required with --analyze-only")
    samples_jsonl = Path(args.samples_jsonl)
    metadata_file = Path(args.metadata_file) if args.metadata_file else samples_jsonl.with_name(samples_jsonl.stem + "_metadata.json")
    results_md = Path(args.results_md) if args.results_md else samples_jsonl.with_name(samples_jsonl.stem + "_results.md")
    summary_json = Path(args.summary_json) if args.summary_json else samples_jsonl.with_name(samples_jsonl.stem + "_summary.json")
    analyze_samples(samples_jsonl, metadata_file, results_md, summary_json, bandwidth_summary, args)
    print(results_md)
    return 0

  artifacts = run_capture(args)
  analyze_samples(artifacts.samples_jsonl, artifacts.metadata_json, artifacts.results_md, artifacts.summary_json, bandwidth_summary, args)
  print(artifacts.results_md)
  return 0


if __name__ == "__main__":
  raise SystemExit(main())



#!/usr/bin/env python3
"""
Stress-test BusinessManager._check_instances with synthetic plugin trees and configs.

Key features:
- Create a fictive plugin filesystem tree from a nested dictionary.
- Generate minimal plugin modules for given signatures.
- Run BusinessManager._check_instances and print timings.

Nested tree format:
{
  "dirA": [
    "file1.txt",
    {"dirB": ["file2.py", {"dirC": ["file3.md"]}]}
  ]
}
"""
from __future__ import annotations

import argparse
import json
import os
import sys
import shutil
import tempfile
import random
from dataclasses import dataclass
from typing import Dict, List, Union, Sequence

Tree = Dict[str, List[Union["Tree", str]]]


@dataclass
class FakeOwner:
  is_supervisor_node: bool = True
  is_secured: bool = False
  evm_network: str | None = None
  __version__: str = "0.0.0"
  runs_in_docker: bool = False
  docker_source: str = "main"

  def set_loop_stage(self, s: str) -> None:
    # Keep minimal but visible if needed.
    return

  def get_pipelines_view(self):
    return {}


def camel_to_snake(s: str) -> str:
  # Mirror Logger.camel_to_snake behavior for deterministic file names.
  if s.isupper():
    return s.lower()
  out = ""
  for i, ch in enumerate(s):
    if ch.isupper() and i > 0:
      out += "_"
    out += ch.lower()
  return out.replace("__", "_")


def signature_to_class_name(signature: str, suffix: str) -> str:
  # _get_module_name_and_class uses name.replace('_','') + suffix
  base = signature.replace("_", "")
  return f"{base}{suffix}"


def ensure_init_py(path: str) -> None:
  os.makedirs(path, exist_ok=True)
  init_path = os.path.join(path, "__init__.py")
  if not os.path.exists(init_path):
    with open(init_path, "w", encoding="utf-8") as f:
      f.write("# auto-generated for test package\n")


def materialize_tree(tree: Tree, root: str) -> None:
  for dirname, items in tree.items():
    dir_path = os.path.join(root, dirname)
    os.makedirs(dir_path, exist_ok=True)
    # Make it package-like to allow importlib to find modules
    ensure_init_py(dir_path)
    for item in items:
      if isinstance(item, dict):
        materialize_tree(item, dir_path)
      elif isinstance(item, str):
        file_path = os.path.join(dir_path, item)
        if not os.path.exists(file_path):
          with open(file_path, "w", encoding="utf-8") as f:
            f.write("# dummy file\n")
      else:
        raise ValueError(f"Unsupported tree item type: {type(item)}")

def count_locations(tree: Tree) -> int:
  count = 0
  for _, items in tree.items():
    count += 1
    for item in items:
      if isinstance(item, dict):
        count += count_locations(item)
  return count


def write_plugin_module(module_dir: str, signature: str, suffix: str,
                        use_base_plugin: bool, use_cv_plugin: bool) -> str:
  file_name = camel_to_snake(signature) + ".py"
  class_name = signature_to_class_name(signature, suffix)
  module_path = os.path.join(module_dir, file_name)
  if use_cv_plugin:
    code = f"""\
from naeural_core.business.base.cv_plugin_executor import CVPluginExecutor

_CONFIG = CVPluginExecutor.CONFIG
__VER__ = "0.0.0"

class {class_name}(CVPluginExecutor):
  def start_thread(self):
    # Avoid starting background threads in tests
    self.thread = None
    return
"""
  elif use_base_plugin:
    code = f"""\
from naeural_core.business.base.base_plugin_biz import BasePluginExecutor

_CONFIG = BasePluginExecutor.CONFIG
__VER__ = "0.0.0"

class {class_name}(BasePluginExecutor):
  def start_thread(self):
    # Avoid starting background threads in tests
    self.thread = None
    return
"""
  else:
    code = f"""\
_CONFIG = {{}}
__VER__ = "0.0.0"

class {class_name}:
  def __init__(self, *args, **kwargs):
    self.cfg_runs_only_on_supervisor_node = False
    self.done_loop = False
    return

  def start_thread(self):
    self.thread = None
    return

  def maybe_update_instance_config(self, **kwargs):
    return

  def __repr__(self):
    return "{class_name}()"
"""
  with open(module_path, "w", encoding="utf-8") as f:
    f.write(code)
  return module_path


def build_fake_plugin_package(base_dir: str, package: str, subdir: str) -> str:
  pkg_root = os.path.join(base_dir, package)
  ensure_init_py(pkg_root)
  if subdir:
    pkg_sub = os.path.join(pkg_root, subdir)
    ensure_init_py(pkg_sub)
    return pkg_sub
  return pkg_root


def generate_signatures(n_unique: int) -> List[str]:
  # Deterministic uppercase signatures to match existing behavior
  return [f"PLG_{i:03d}" for i in range(1, n_unique + 1)]


def build_instances_config(signatures: List[str], total_instances: int) -> list:
  from naeural_core import constants as ct
  plugins = []
  if not signatures:
    return plugins
  per_sig = total_instances // len(signatures)
  remainder = total_instances % len(signatures)

  idx = 0
  for sig in signatures:
    n = per_sig + (1 if idx < remainder else 0)
    idx += 1
    instances = []
    for j in range(n):
      instance_id = f"{sig}_INST_{j+1:03d}"
      instances.append({
        ct.CONFIG_INSTANCE.K_INSTANCE_ID: instance_id,
        "DISABLED": False,
      })
    plugins.append({
      ct.CONFIG_PLUGIN.K_SIGNATURE: sig,
      ct.CONFIG_PLUGIN.K_INSTANCES: instances,
    })
  random.shuffle(plugins)
  return plugins


def build_streams_config(plugins: list, pipeline_name: str = "pipeline_test") -> dict:
  from naeural_core import constants as ct
  return {
    pipeline_name: {
      ct.CONFIG_STREAM.K_PLUGINS: plugins,
      ct.CONFIG_STREAM.K_INITIATOR_ID: "test_initiator",
      ct.CONFIG_STREAM.K_INITIATOR_ADDR: "local",
      ct.CONFIG_STREAM.K_MODIFIED_BY_ID: "test_modifier",
      ct.CONFIG_STREAM.K_MODIFIED_BY_ADDR: "local",
      ct.CONFIG_STREAM.K_SESSION_ID: "session_test",
    }
  }


def load_tree(path: str | None) -> Tree | None:
  if path is None:
    return None
  with open(path, "r", encoding="utf-8") as f:
    return json.load(f)

def parse_signatures_csv(csv_value: str | None) -> List[str]:
  if not csv_value:
    return []
  return [x.strip() for x in csv_value.split(",") if x.strip()]

def _build_file_names(count: int, exts: Sequence[str], prefix: str) -> List[str]:
  names: List[str] = []
  for i in range(count):
    ext = exts[i % len(exts)] if exts else ""
    names.append(f"{prefix}_{i:03d}{ext}")
  return names


def generate_tree(depth: int, dirs_per_level: int, files_per_dir: int,
                  file_exts: Sequence[str], full_breadth: bool,
                  rng: random.Random, level: int = 0) -> Tree:
  if depth <= 0:
    return {}

  if full_breadth:
    n_dirs = dirs_per_level
    n_files = files_per_dir
  else:
    n_dirs = rng.randint(0, dirs_per_level)
    n_files = rng.randint(0, files_per_dir)

  dir_name = f"dir_{level}_{rng.randint(0, 9999):04d}"
  items: List[Union[Tree, str]] = []
  items.extend(_build_file_names(n_files, file_exts, f"file_{level}"))

  if depth > 1:
    for _ in range(n_dirs):
      items.append(generate_tree(depth - 1, dirs_per_level, files_per_dir,
                                 file_exts, full_breadth, rng, level + 1))

  return {dir_name: items}


def main() -> int:
  parser = argparse.ArgumentParser(description="Test BusinessManager._check_instances")
  parser.add_argument("--total-fake-instances", type=int, default=20)
  parser.add_argument("--total-real-instances", type=int, default=120)
  parser.add_argument("--unique-signatures", type=int, default=15)
  parser.add_argument("--fake-fs", type=str, default=None,
                      help="Path to JSON file describing a fake filesystem tree")
  parser.add_argument("--gen-tree", action="store_true",
                      help="Generate a fake filesystem tree instead of loading JSON")
  parser.add_argument("--tree-depth", type=int, default=6)
  parser.add_argument("--dirs-per-level", type=int, default=5)
  parser.add_argument("--files-per-dir", type=int, default=3)
  # parser.add_argument("--file-exts", type=str, default=".py,.txt,.md")
  parser.add_argument("--file-exts", type=str, default=".py")
  parser.add_argument("--full-breadth", action="store_true",
                      help="If set, each directory has exactly files_per_dir and dirs_per_level")
  parser.add_argument("--seed", type=int, default=1337)
  parser.add_argument("--use-base-plugin", action="store_true",
                      help="Generate plugins that subclass BasePluginExecutor")
  parser.add_argument("--use-cv-plugin", action="store_true",
                      help="Generate plugins that subclass CVPluginExecutor")
  parser.add_argument("--real-signatures", type=str, default="",
                      help="Comma-separated list of real plugin signatures to include")
  parser.add_argument("--include-real-plugins", action="store_true",
                      help="Include real plugins in search paths")
  parser.add_argument("--no-threads", action="store_true",
                      help="Disable plugin threads by monkeypatching BasePluginExecutor.start_thread")
  parser.add_argument("--keep", action="store_true", help="Keep generated fake filesystem on disk")
  args = parser.parse_args()

  repo_root = os.path.abspath(os.path.join(os.path.dirname(__file__), "..", ".."))
  # repo_root = os.path.join(repo_root, "naeural_core")
  # print(f"repo_root: {repo_root}")
  # exit(-1)

  # Drop any previously imported naeural_core to force local reload
  for mod in list(sys.modules.keys()):
    if mod == "naeural_core" or mod.startswith("naeural_core."):
      del sys.modules[mod]


  # Ensure local repo has priority over any installed package
  if repo_root not in sys.path:
    sys.path.insert(0, repo_root)

  from naeural_core import constants as ct
  from naeural_core import Logger
  from naeural_core.business.business_manager import BusinessManager

  # Create temp base and fake package structure
  base_dir = tempfile.mkdtemp(prefix="bm_check_instances_")
  package_name = "fake_biz_plugins"
  subdir = "biz"

  # Materialize optional fake tree to stress filesystem traversal
  tree = load_tree(args.fake_fs)
  if tree is None and args.gen_tree:
    exts = [x for x in args.file_exts.split(",") if x]
    rng = random.Random(args.seed)
    tree = generate_tree(args.tree_depth, args.dirs_per_level, args.files_per_dir,
                         exts, args.full_breadth, rng)
  if tree:
    materialize_tree(tree, base_dir)
    print(f"Generated fake filesystem locations: {count_locations(tree)}")

  plugin_dir = build_fake_plugin_package(base_dir, package_name, subdir)

  if args.use_cv_plugin and args.use_base_plugin:
    raise ValueError("Use only one of --use-base-plugin or --use-cv-plugin")

  if args.use_base_plugin or args.use_cv_plugin or args.include_real_plugins:
    # Pre-import serving modules to mirror production import order and avoid cold-start cycles.
    import naeural_core.serving  # noqa: F401
    import naeural_core.serving.ai_engines  # noqa: F401

  if args.no_threads:
    from naeural_core.business.base.base_plugin_biz import BasePluginExecutor
    def _noop_start_thread(self):
      self.thread = None
      return
    BasePluginExecutor.start_thread = _noop_start_thread

  # Create fake plugin modules
  fake_signatures = generate_signatures(args.unique_signatures)
  for sig in fake_signatures:
    write_plugin_module(plugin_dir, sig, ct.PLUGIN_SEARCH.SUFFIX_BIZ_PLUGINS,
                        use_base_plugin=args.use_base_plugin,
                        use_cv_plugin=args.use_cv_plugin)
  print(f"Generated fake plugin modules: {len(fake_signatures)}")

  real_signatures = parse_signatures_csv(args.real_signatures)
  if args.include_real_plugins and not real_signatures:
    raise ValueError("--include-real-plugins requires --real-signatures")

  if args.include_real_plugins:
    all_signatures = []
    for sig in fake_signatures + real_signatures:
      if sig not in all_signatures:
        all_signatures.append(sig)
  else:
    all_signatures = list(fake_signatures)

  # Prepare environment: ensure both repo root and temp base are importable
  sys.path.insert(0, repo_root)
  sys.path.insert(0, base_dir)

  # For local filesystem scanning, _get_plugin_by_name uses os.walk on relative paths.
  # We temporarily switch cwd to base_dir so it can discover fake_biz_plugins.
  orig_cwd = os.getcwd()
  os.chdir(base_dir)

  # Patch plugin search locations to point to our temp package (and optionally real locations)
  orig_loc = ct.PLUGIN_SEARCH.LOC_BIZ_PLUGINS
  orig_safe = ct.PLUGIN_SEARCH.SAFE_BIZ_PLUGINS
  try:
    if args.include_real_plugins:
      ct.PLUGIN_SEARCH.LOC_BIZ_PLUGINS = orig_loc + [f"{package_name}.{subdir}"]
      ct.PLUGIN_SEARCH.SAFE_BIZ_PLUGINS = orig_safe
    else:
      ct.PLUGIN_SEARCH.LOC_BIZ_PLUGINS = [f"{package_name}.{subdir}"]
      ct.PLUGIN_SEARCH.SAFE_BIZ_PLUGINS = []

    log = Logger("BIZM_TEST", DEBUG=False)
    owner = FakeOwner()
    class _DummyBC:
      address = ""
    class _DummyR1FS:
      pass
    def _dummy_main_loop_resolution():
      return 0.1
    def _dummy_save_config(**kwargs):
      return None
    class _DummyNetMon:
      pass
    shmem = {
      ct.BLOCKCHAIN_MANAGER: _DummyBC(),
      ct.R1FS_ENGINE: _DummyR1FS(),
      ct.CALLBACKS.MAIN_LOOP_RESOLUTION_CALLBACK: _dummy_main_loop_resolution,
      ct.CALLBACKS.INSTANCE_CONFIG_SAVER_CALLBACK: _dummy_save_config,
      "network_monitor": _DummyNetMon(),
    }
    log.config_data['BM_CACHE_VALIDATORS'] = True
    log.config_data['BM_CACHE_CONFIG_HANDLERS'] = True
    bm = BusinessManager(log=log, owner=owner, shmem=shmem, run_on_threads=True)

    if args.include_real_plugins:
      plugins_cfg = []
      plugins_cfg += build_instances_config(fake_signatures, args.total_fake_instances)
      plugins_cfg += build_instances_config(real_signatures, args.total_real_instances)
      streams_cfg = build_streams_config(plugins_cfg)
    else:
      plugins_cfg = build_instances_config(fake_signatures, args.total_fake_instances)
      streams_cfg = build_streams_config(plugins_cfg)

    bm._dct_config_streams = streams_cfg
    bm._check_instances()

  finally:
    ct.PLUGIN_SEARCH.LOC_BIZ_PLUGINS = orig_loc
    ct.PLUGIN_SEARCH.SAFE_BIZ_PLUGINS = orig_safe
    os.chdir(orig_cwd)
    if not args.keep:
      shutil.rmtree(base_dir, ignore_errors=True)

  return 0


if __name__ == "__main__":
  raise SystemExit(main())



from ultralytics import YOLO
import torch as th
import json
import os
from pathlib import Path
import shutil
import gc
import torchvision

from naeural_core import Logger

def yolo_models(
    log, classnames=None, classnames_donate_from=None,
):
  model_cfgs = {
    # '8n': {
    #   'yaml': 'yolov8n.yaml',
    #   'weights': 'yolov8n.pt',
    # },
    # '8s': {
    #   'yaml': 'yolov8s.yaml',
    #   'weights': 'yolov8s.pt',
    # },
    # '8l': {
    #   'yaml': 'yolov8l.yaml',
    #   'weights': 'yolov8l.pt',
    # },
    '11s_drone_v1': {
      'yaml': 'yolo11s.yaml',
      # 'weights': r'C:\repos\HyFy_E2\xperimental\lpd\drone_v1_y11s.pt',
      'weights': r'C:\repos\HyFy_E2\xperimental\lpd\runs\detect\train42\weights\best.pt',
    }
  }

  sizes = {
    # '8n': [
    #   [448, 640],
    # ],
    # '8s': [
    #   [640, 1152],
    # ],
    # '8l': [
    #   [640, 896],
    # ],
    # '8x': [
    #   [1152, 2048]
    # ],
    '11s_drone_v1': [
      [896, 1280],
    ],
  }
  if classnames_donate_from is None:
    classnames_donate_from = os.path.join(log.get_models_folder(), '20230723_y8l_nms.ths')
  # endif classnames_donate_from is None
  if classnames is None:
    extra_files = {'config.txt': ''}
    model = th.jit.load(classnames_donate_from, map_location='cpu', _extra_files=extra_files)
    classnames = json.loads(extra_files['config.txt'])['names']
    model = None
    gc.collect()
  # endif classnames is None

  for model_name in sizes.keys():
    gc.collect()
    th.cuda.empty_cache()

    # model = YOLO(model_cfgs[model_name]['yaml']).load(model_cfgs[model_name]['weights'])
    model = YOLO(model_cfgs[model_name]['weights'])  # to avoid loading optimizer etc.
    device = 'cuda:0'
    model = model.to(device)
    format = "torchscript"

    for imgsz in sizes[model_name]:
      log.P(f'Exporting y{model_name} with imgsz={imgsz}')
      export_kwargs = {
        'format': format,
        'imgsz': imgsz,
        'names': classnames,
      }
      pt_path = f'y{model_name}_{imgsz[0]}x{imgsz[1]}.torchscript'
      setattr(model.model, 'pt_path', pt_path)
      setattr(model.model, 'names', classnames)
      if os.path.exists(pt_path):
        log.P(f"Extracting config from torchscript {pt_path}")
        extra_files = {'config.txt': ''}
        _ = th.jit.load(pt_path, map_location='cpu', _extra_files=extra_files)
        file_config = json.loads(extra_files['config.txt'])['names']
        export_kwargs = {
          **file_config,
          **export_kwargs
        }
        gc.collect()
      # endif os.path.exists(pt_path)
      yield (model, export_kwargs)

    # endfor imgsz
    # Delete the model and make sure we've freed the memory.
  # endfor model_name

if __name__ == '__main__':
  log = Logger('Y8', base_folder='.', app_folder='_local_cache')
  models_folder = log.get_models_folder()
  for model, export_kwargs in yolo_models(log=log):
    exported = model.export(**export_kwargs)
    shutil.move(model.model.pt_path, os.path.join(models_folder, model.model.pt_path))
    del model
    gc.collect()




from naeural_core.business.mixins_base.tunnel_engine_mixin import _TunnelEngineMixin


CLOUDFLARE_DEFAULT_PARAMETERS = {
  "CLOUDFLARE_TOKEN": None,
  "CLOUDFLARE_PROTOCOL": "http",
}


class _CloudflareMixinPlugin(_TunnelEngineMixin):
  """
  A plugin which exposes all of its methods marked with @endpoint through
  fastapi as http endpoints, and further tunnels traffic to this interface
  via cloudflare.

  The @endpoint methods can be triggered via http requests on the web server
  and will be processed as part of the business plugin loop.
  """
  """CLOUDFLARE UTILS METHODS"""
  if True:
    def _get_cloudflare_start_command(self):
      token = self.get_cloudflare_token()
      protocol = self.get_cloudflare_protocol()
      if token is not None:
        return f"cloudflared tunnel --no-autoupdate run --token {token} --url {protocol}://127.0.0.1:{self.port}"
      return f"cloudflared tunnel --no-autoupdate --url {protocol}://127.0.0.1:{self.port}"
  """END CLOUDFLARE UTILS METHODS"""

  """RETRIEVE CLOUDFLARE SPECIFIC CONFIGURATION_PARAMETERS"""
  if True:
    def get_cloudflare_token(self):
      """
      Retrieve the Cloudflare token from the configuration parameters.
      If not set, it returns None.
      """
      # Here .get is used because init_endpoints goes through every method to find all
      # endpoints. Because this method is used by the property method app_url_cloudflare it was
      # called during the search
      return self.get_tunnel_engine_parameters().get("CLOUDFLARE_TOKEN")
    
    def get_cloudflare_protocol(self):
      """
      Retrieve the Cloudflare protocol from the configuration parameters.
      If not set, it returns "http".
      """
      protocol = self.get_tunnel_engine_parameters().get("CLOUDFLARE_PROTOCOL")
      if protocol is None:
        protocol = "http"
      return protocol
  """END RETRIEVE CLOUDFLARE SPECIFIC CONFIGURATION_PARAMETERS"""

  """BASE CLASS METHODS"""
  if True:
    @property
    def app_url_cloudflare(self):
      return None if self.get_cloudflare_token() is not None or not hasattr(self,
                                                                            '_CloudflareMixinPlugin__app_url') else self.__app_url
    def get_default_tunnel_engine_parameters_cloudflare(self):
      return CLOUDFLARE_DEFAULT_PARAMETERS

    def reset_tunnel_engine_cloudflare(self):
      """
      Reset the tunnel engine by stopping any existing tunnel and clearing the configuration.
      """
      super(_CloudflareMixinPlugin, self).reset_tunnel_engine()
      self.__app_url = None
      return

    def maybe_init_tunnel_engine_cloudflare(self):
      super(_CloudflareMixinPlugin, self).maybe_init_tunnel_engine()
      return

    def maybe_start_tunnel_engine_cloudflare(self):
      super(_CloudflareMixinPlugin, self).maybe_start_tunnel_engine()
      return

    def maybe_stop_tunnel_engine_cloudflare(self):
      super(_CloudflareMixinPlugin, self).maybe_stop_tunnel_engine()
      return

    def get_setup_commands_cloudflare(self):
      super(_CloudflareMixinPlugin, self).get_setup_commands()
      return

    def get_start_commands_cloudflare(self):
      super_start_commands = super(_CloudflareMixinPlugin, self).get_start_commands()

      if self.cfg_tunnel_engine_enabled:
        super_start_commands = super_start_commands + [self._get_cloudflare_start_command()]
      # endif tunnel engine enabled
      return super_start_commands

    def check_valid_tunnel_engine_config_cloudflare(self):
      """
      Check if the tunnel engine configuration is valid.
      If the Cloudflare token is not set, it raises an error.
      """
      is_valid, msg = True, None
      token = self.get_cloudflare_token()
      if token is None or token == "":
        msg = "Cloudflare token is not set."
        msg += "Please set the `CLOUDFLARE_TOKEN` parameter in your configuration."
        is_valid = False
      # endif token is None
      return is_valid, msg

    def on_log_handler_cloudflare(self, text, key=None):
      """
      Handle log messages from the Cloudflare tunnel.
      This method can be overridden in subclasses to handle logs differently.
      """
      super(_CloudflareMixinPlugin, self).on_log_handler(text, key)
      if hasattr(self, '_CloudflareMixinPlugin__app_url') and self.__app_url is not None:
        # URL already set, no need to process further
        return
      # Define a regular expression pattern to match the URL format
      url_pattern = r'https://[a-z0-9-]+\.trycloudflare\.com'
      match = self.re.search(url_pattern, text)

      if match:
        self.__app_url = match.group(0)
        self.P(f"Cloudflare tunnel started successfully on: {self.__app_url}", color="green")
      return
  """END BASE CLASS METHODS"""




import builtins

from naeural_core.constants import SUPERVISOR_MIN_AVAIL_PRC

from naeural_core.business.default.web_app.naeural_fast_api_web_app import NaeuralFastApiWebApp

_CONFIG = {
  **NaeuralFastApiWebApp.CONFIG,

  # Only for the debug plugin
  "ASSETS": "extensions/business/xperimental",
  "DEBUG_MODE": False,
  "REQUEST_TIMEOUT": 5,

  'VALIDATION_RULES': {
    **NaeuralFastApiWebApp.CONFIG['VALIDATION_RULES'],
  },
}


class NaeuralFastApiDebugPlugin(NaeuralFastApiWebApp):
  """
  Debug plugin class for the Naeural Fast API Web App interface.
  """
  CONFIG = _CONFIG
  @NaeuralFastApiWebApp.endpoint(method="post")
  def test_malicious_file(self, subdir: str = None, filename: str = None):
    res_path = self.diskapi_save_file_to_output(
      subdir=subdir or './_bin',
      filename=filename or 'text.json',
      data="asjegbisudhguisfogbdg"
    )
    return {
      "res_path": res_path,
    }

  @NaeuralFastApiWebApp.endpoint(method='post')
  def test_request_timeout(self, timeout_s: int = 10):
    self.P(f"Starting sleep for {timeout_s} seconds...", color='y')
    self.sleep(timeout_s)
    self.P(f"Finished sleep for {timeout_s} seconds.", color='g')
    return {
      "status": "success",
      "slept_for": timeout_s
    }

  @NaeuralFastApiWebApp.endpoint(method='post')
  def test_llama_cpp(self):
    response = {}
    try:
      from llama_cpp import Llama
      response['import'] = True

      cache_dir = self.log.get_models_folder()
      model_params = {
        'n_ctx': 4096,
        'seed': 42,
        'n_batch': 512,
      }

      model = Llama.from_pretrained(
        repo_id="mradermacher/DatA-SQL-1.5B-i1-GGUF",
        filename="DatA-SQL-1.5B.i1-Q4_K_M.gguf",
        cache_dir=cache_dir,
        **model_params,
      )
      response['model_init'] = True

      messages = [
        {
          "role": "system",
          "content": "You are a well known baker."
        },
        {
          "role": "user",
          "content": "Name 3 french pastries and describe why they piss you off."
        }
      ]
      result = model.create_chat_completion(
        messages=messages,
        temperature=0.5
      )
      response["model_inference"] = True
      response["result"] = result

      del model

      response['success'] = True
    except Exception as e:
      response['error'] = str(e)
    return response


    return
  # @NaeuralFastApiWebApp.endpoint(method='post')
  # def command_node(self, node_addr: str):
  #   self.cmdapi_start_pipeline(
  #     node_address=node_addr,
  #     config={
  #       "NAME": "empty_pipeline",
  #       "TYPE": "VOID"
  #     }
  #   )
  #   return

  @NaeuralFastApiWebApp.endpoint(method='post')
  def post_stuff(self, stuff1: str = "", stuff2: str = ""):
    """
    Example of a POST endpoint that receives two strings and returns them in a dictionary.
    """
    self.P(f"Received `{stuff1 = }` and `{stuff2 = }`", color='b')
    stuff1 = stuff1 or {"empty": True}
    stuff2 = stuff2 or {"empty": True}
    return {
      'Status': 'Success',
      'Stuff1': stuff1,
      'Stuff2': stuff2
    }

  @NaeuralFastApiWebApp.endpoint(method='post')
  def post_stuff_with_extras(self, stuff1: str = "", stuff2: str = "alabala", **kwargs):
    """
    Example of a POST endpoint that receives two strings and returns them in a dictionary.
    """
    self.P(f"Received `{stuff1 = }` and `{stuff2 = }` and ", color='b')
    stuff1 = stuff1 or {"empty": True}
    stuff2 = stuff2 or {"empty": True}
    return {
      'Status': 'Success',
      'Stuff1': stuff1,
      'Stuff2': stuff2,
      'Extra': self.deepcopy(kwargs)
    }

  @NaeuralFastApiWebApp.endpoint(method='post')
  def post_required_with_extras(self, param1: str, param2: int, **kwargs):
    self.P(f"Received `{param1 = }` | `{param2 = }` and extras = {kwargs}", color='b')
    return {
      'Status': 'Success',
      'Param1': param1,
      'Param2': param2,
      'Extra': self.deepcopy(kwargs)
    }

  @NaeuralFastApiWebApp.endpoint(method='post')
  def post_required_without_extras(self, param1: str, param2: int):
    self.P(f"Received `{param1 = }` | `{param2 = }`", color='b')
    return {
      'Status': 'Success',
      'Param1': param1,
      'Param2': param2,
    }

  def self_assessment(self):
    """
    Perform self-assessment throughout the epoch to know the local availability so far and
    to predict the final availability of the node at the end of the epoch.
    """
    SUPERVISOR_MIN_AVAIL_PRC = 0.98

    total_seconds_availability, total_seconds_from_start = self.netmon.epoch_manager.get_current_epoch_availability(
      return_absolute=True,
      return_max=True
    )
    total_epoch_seconds = self.netmon.epoch_manager.epoch_length
    prc_node_availability = total_seconds_availability / total_epoch_seconds
    prc_max_availability = total_seconds_from_start / total_epoch_seconds
    prc_missed_availability = prc_max_availability - prc_node_availability
    prc_predicted_availability = 1 - prc_missed_availability
    will_participate = prc_predicted_availability >= SUPERVISOR_MIN_AVAIL_PRC
    comparing_str = f"{'>=' if will_participate else '<='} {SUPERVISOR_MIN_AVAIL_PRC:.2%}"
    comparing_str += f" => {'will' if will_participate else 'will not'} participate in the sync process."
    log_str = f"Current self-assessment:\n"
    log_str += f"\tNode current availability: {prc_node_availability:.2%}\n"
    log_str += f"\tPassed from epoch: {prc_max_availability:.2%}\n"
    log_str += f"\tMissed availability so far: {prc_missed_availability:.2%}\n"
    log_str += f"\tPredicted availability at the end of epoch: {prc_predicted_availability:.2%}{comparing_str}\n"
    self.P(log_str, color='g')
    return


  @NaeuralFastApiWebApp.endpoint(method='get', require_token=True)
  def get_stuff_with_token(self, token: str, stuff: str = ""):
    self.self_assessment()
    if token not in ['123', 'alabala']:
      return "Unauthorized token"
    stuff = stuff or {"empty": True}
    return {
      'Status': 'Success',
      'Stuff': stuff
    }

  @NaeuralFastApiWebApp.endpoint(method='get')
  def get_stuff_with_extras(self, stuff: str = None, **kwargs):
    self.self_assessment()
    stuff = stuff or {"empty": True}
    self.P(f"Received extras: {kwargs}", color='b')

    return {
      'Status': 'Success',
      'Stuff': stuff,
      'Extra': self.deepcopy(kwargs)
    }

  @NaeuralFastApiWebApp.endpoint(method='post')
  def trigger_exception(self, exc_type: str = "ValueError", message: str = "debug crash"):
    """
    Raise a user-selected exception to exercise FastAPI/IPC error handling paths.

    Predefined exception names you can use: ValueError, RuntimeError, KeyError, TypeError, AssertionError, ZeroDivisionError.
    """
    predefined = {
      "ValueError": ValueError,
      "RuntimeError": RuntimeError,
      "KeyError": KeyError,
      "TypeError": TypeError,
      "AssertionError": AssertionError,
      "ZeroDivisionError": ZeroDivisionError,
    }
    exc_cls = predefined.get(exc_type) or getattr(builtins, exc_type, RuntimeError)
    raise exc_cls(message)

  @NaeuralFastApiWebApp.endpoint(method='get')
  def get_stuff_without_token(self, stuff: str = ""):
    stuff = stuff or {"empty": True}
    return {
      'Status': 'Success',
      'Stuff': stuff
    }

